#!/usr/bin/env python
# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

import SimpleITK as sitk

from platipy.dicom.io.nifti_to_series import convert_nifti_to_dicom_series


@click.command()
@click.option(
    "--dcm",
    "-d",
    required=True,
    help="Reference DICOM folder containing series from which header tags will be copied",
)
@click.option("--image", "-i", required=True, help="Nifti Image file to convert to Dicom series")
@click.option(
    "--tag", "-t", multiple=True, help="Override Dicom Tags by providing key:value pairs"
)
@click.option(
    "--output_directory",
    "-o",
    help="Directory in which Dicom series files will be generated",
    default=".",
    show_default=True,
)
def click_command(dcm, image, tag, output_directory):
    """
    Convert a Nifti image to a Dicom image series
    """

    tags = [t.split(":") for t in tag]
    convert_nifti_to_dicom_series(
        sitk.ReadImage(image), dcm, tag_overrides=tags, output_directory=output_directory
    )


if __name__ == "__main__":
    click_command()  # pylint: disable=no-value-for-parameter



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import SimpleITK as sitk

from platipy.imaging.utils.crop import label_to_roi, crop_to_roi


def compute_volume(label):
    """Computes the volume in cubic centimetres

    Args:
        label (SimpleITK.Image): A binary mask.

    Returns:
        float: The volume (in cubic centimetres)
    """

    return sitk.GetArrayFromImage(label).sum() * np.product(label.GetSpacing()) / 1000


def compute_surface_dsc(label_a, label_b, tau=3.0):
    """Compute Surface Dice

    From: Nikolov S et al. Clinically Applicable Segmentation of Head and Neck Anatomy for
    Radiotherapy: Deep Learning Algorithm Development and Validation Study J Med Internet Res
    2021;23(7):e26151, DOI: 10.2196/26151

    Args:
        label_a (sitk.Image): A mask to compare
        label_b (sitk.Image): Another mask to compare
        tau (float): Accepted deviation between contours (in mm)

    Returns:
        float: The Surface DSC between the two labels
    """

    binary_contour_filter = sitk.BinaryContourImageFilter()
    binary_contour_filter.FullyConnectedOn()
    a_contour = binary_contour_filter.Execute(label_a)
    b_contour = binary_contour_filter.Execute(label_b)

    dist_to_a = sitk.SignedMaurerDistanceMap(
        a_contour, useImageSpacing=True, squaredDistance=False
    )

    dist_to_b = sitk.SignedMaurerDistanceMap(
        b_contour, useImageSpacing=True, squaredDistance=False
    )

    b_intersection = sitk.GetArrayFromImage(b_contour * (dist_to_a <= tau)).sum()
    a_intersection = sitk.GetArrayFromImage(a_contour * (dist_to_b <= tau)).sum()

    surface_sum = (
        sitk.GetArrayFromImage(a_contour).sum()
        + sitk.GetArrayFromImage(b_contour).sum()
    )

    return (b_intersection + a_intersection) / surface_sum


def compute_surface_metrics(label_a, label_b, verbose=False):
    """Compute surface distance metrics between two labels. Surface metrics computed are:
    hausdorffDistance, meanSurfaceDistance, medianSurfaceDistance, maximumSurfaceDistance,
    sigmaSurfaceDistance, surfaceDSC

    Args:
        label_a (sitk.Image): A mask to compare
        label_b (sitk.Image): Another mask to compare
        verbose (bool, optional): Whether to print verbose output. Defaults to False.

    Returns:
        dict: Dictionary object containing surface distance metrics
    """

    hausdorff_distance = sitk.HausdorffDistanceImageFilter()
    hausdorff_distance.Execute(label_a, label_b)
    hd = hausdorff_distance.GetHausdorffDistance()

    mean_sd_list = []
    max_sd_list = []
    std_sd_list = []
    median_sd_list = []
    num_points = []
    for (la, lb) in ((label_a, label_b), (label_b, label_a)):

        label_intensity_stat = sitk.LabelIntensityStatisticsImageFilter()
        reference_distance_map = sitk.Abs(
            sitk.SignedMaurerDistanceMap(
                la, squaredDistance=False, useImageSpacing=True
            )
        )
        moving_label_contour = sitk.LabelContour(lb)
        label_intensity_stat.Execute(moving_label_contour, reference_distance_map)

        mean_sd_list.append(label_intensity_stat.GetMean(1))
        max_sd_list.append(label_intensity_stat.GetMaximum(1))
        std_sd_list.append(label_intensity_stat.GetStandardDeviation(1))
        median_sd_list.append(label_intensity_stat.GetMedian(1))

        num_points.append(label_intensity_stat.GetNumberOfPixels(1))

    if verbose:
        print("        Boundary points:  {0}  {1}".format(num_points[0], num_points[1]))

    mean_surf_dist = np.dot(mean_sd_list, num_points) / np.sum(num_points)
    max_surf_dist = np.max(max_sd_list)
    std_surf_dist = np.sqrt(
        np.dot(
            num_points,
            np.add(
                np.square(std_sd_list),
                np.square(np.subtract(mean_sd_list, mean_surf_dist)),
            ),
        )
    )
    median_surf_dist = np.mean(median_sd_list)

    result = {}
    result["hausdorffDistance"] = hd
    result["meanSurfaceDistance"] = mean_surf_dist
    result["medianSurfaceDistance"] = median_surf_dist
    result["maximumSurfaceDistance"] = max_surf_dist
    result["sigmaSurfaceDistance"] = std_surf_dist
    result["surfaceDSC"] = compute_surface_dsc(label_a, label_b)

    return result


def compute_volume_metrics(label_a, label_b):
    """Compute volume metrics between two labels. Volume metrics computed are:
    DSC, volumeOverlap fractionOverlap truePositiveFraction trueNegativeFraction
    falsePositiveFraction falseNegativeFraction

    Args:
        label_a (sitk.Image): A mask to compare
        label_b (sitk.Image): Another mask to compare

    Returns:
        dict: Dictionary object containing volume metrics
    """

    arr_a = sitk.GetArrayFromImage(label_a).astype(bool)
    arr_b = sitk.GetArrayFromImage(label_b).astype(bool)

    arr_intersection = arr_a & arr_b
    arr_union = arr_a | arr_b

    voxel_volume = np.product(label_a.GetSpacing()) / 1000.0  # Conversion to cm^3

    # 2|A & B|/(|A|+|B|)
    dsc = (2.0 * arr_intersection.sum()) / (arr_a.sum() + arr_b.sum())

    #  |A & B|/|A | B|
    frac_overlap = arr_intersection.sum() / arr_union.sum().astype(float)
    vol_overlap = arr_intersection.sum() * voxel_volume

    true_pos = arr_intersection.sum()
    true_neg = (np.invert(arr_a) & np.invert(arr_b)).sum()
    false_pos = arr_b.sum() - true_pos
    false_neg = arr_a.sum() - true_pos

    true_pos_frac = (1.0 * true_pos) / (true_pos + false_neg)
    true_neg_frac = (1.0 * true_neg) / (true_neg + false_pos)
    false_pos_frac = (1.0 * false_pos) / (true_neg + false_pos)
    false_neg_frac = (1.0 * false_neg) / (true_pos + false_neg)

    result = {}
    result["DSC"] = dsc
    result["volumeOverlap"] = vol_overlap
    result["fractionOverlap"] = frac_overlap
    result["truePositiveFraction"] = true_pos_frac
    result["trueNegativeFraction"] = true_neg_frac
    result["falsePositiveFraction"] = false_pos_frac
    result["falseNegativeFraction"] = false_neg_frac

    return result


def compute_metric_dsc(label_a, label_b, auto_crop=True):
    """Compute the Dice Similarity Coefficient between two labels

    Args:
        label_a (sitk.Image): A mask to compare
        label_b (sitk.Image): Another mask to compare

    Returns:
        float: The Dice Similarity Coefficient
    """
    if auto_crop:
        largest_region = (label_a + label_b) > 0
        crop_box_size, crop_box_index = label_to_roi(largest_region)

        label_a = crop_to_roi(label_a, size=crop_box_size, index=crop_box_index)
        label_b = crop_to_roi(label_b, size=crop_box_size, index=crop_box_index)

    arr_a = sitk.GetArrayFromImage(label_a).astype(bool)
    arr_b = sitk.GetArrayFromImage(label_b).astype(bool)
    return 2 * ((arr_a & arr_b).sum()) / (arr_a.sum() + arr_b.sum())


def compute_metric_specificity(label_a, label_b, auto_crop=True):
    """Compute the specificity between two labels

    Args:
        label_a (sitk.Image): A mask to compare
        label_b (sitk.Image): Another mask to compare

    Returns:
        float: The specificity between the two labels
    """
    if auto_crop:
        largest_region = (label_a + label_b) > 0
        crop_box_size, crop_box_index = label_to_roi(largest_region)

        label_a = crop_to_roi(label_a, size=crop_box_size, index=crop_box_index)
        label_b = crop_to_roi(label_b, size=crop_box_size, index=crop_box_index)

    arr_a = sitk.GetArrayFromImage(label_a).astype(bool)
    arr_b = sitk.GetArrayFromImage(label_b).astype(bool)

    arr_intersection = arr_a & arr_b

    true_pos = arr_intersection.sum()
    true_neg = (np.invert(arr_a) & np.invert(arr_b)).sum()
    false_pos = arr_b.sum() - true_pos

    return float((1.0 * true_neg) / (true_neg + false_pos))


def compute_metric_sensitivity(label_a, label_b, auto_crop=True):
    """Compute the sensitivity between two labels

    Args:
        label_a (sitk.Image): A mask to compare
        label_b (sitk.Image): Another mask to compare

    Returns:
        float: The sensitivity between the two labels
    """
    if auto_crop:
        largest_region = (label_a + label_b) > 0
        crop_box_size, crop_box_index = label_to_roi(largest_region)

        label_a = crop_to_roi(label_a, size=crop_box_size, index=crop_box_index)
        label_b = crop_to_roi(label_b, size=crop_box_size, index=crop_box_index)

    arr_a = sitk.GetArrayFromImage(label_a).astype(bool)
    arr_b = sitk.GetArrayFromImage(label_b).astype(bool)

    arr_intersection = arr_a & arr_b

    true_pos = arr_intersection.sum()
    false_neg = arr_a.sum() - true_pos

    return float((1.0 * true_pos) / (true_pos + false_neg))


def compute_metric_masd(label_a, label_b, auto_crop=True):
    """Compute the mean absolute distance between two labels

    Args:
        label_a (sitk.Image): A mask to compare
        label_b (sitk.Image): Another mask to compare

    Returns:
        float: The mean absolute surface distance
    """
    if auto_crop:
        largest_region = (label_a + label_b) > 0
        crop_box_size, crop_box_index = label_to_roi(largest_region)

        label_a = crop_to_roi(label_a, size=crop_box_size, index=crop_box_index)
        label_b = crop_to_roi(label_b, size=crop_box_size, index=crop_box_index)

    if (
        sitk.GetArrayViewFromImage(label_a).sum() == 0
        or sitk.GetArrayViewFromImage(label_b).sum() == 0
    ):
        return np.nan

    mean_sd_list = []
    num_points = []
    for (la, lb) in ((label_a, label_b), (label_b, label_a)):

        label_intensity_stat = sitk.LabelIntensityStatisticsImageFilter()
        reference_distance_map = sitk.Abs(
            sitk.SignedMaurerDistanceMap(
                la, squaredDistance=False, useImageSpacing=True
            )
        )
        moving_label_contour = sitk.LabelContour(lb)
        label_intensity_stat.Execute(moving_label_contour, reference_distance_map)

        mean_sd_list.append(label_intensity_stat.GetMean(1))
        num_points.append(label_intensity_stat.GetNumberOfPixels(1))

    mean_surf_dist = np.dot(mean_sd_list, num_points) / np.sum(num_points)
    return float(mean_surf_dist)


def compute_metric_hd(label_a, label_b, auto_crop=True):
    """Compute the Hausdorff distance between two labels

    Args:
        label_a (sitk.Image): A mask to compare
        label_b (sitk.Image): Another mask to compare

    Returns:
        float: The maximum Hausdorff distance
    """

    if auto_crop:
        largest_region = (label_a + label_b) > 0
        crop_box_size, crop_box_index = label_to_roi(largest_region)

        label_a = crop_to_roi(label_a, size=crop_box_size, index=crop_box_index)
        label_b = crop_to_roi(label_b, size=crop_box_size, index=crop_box_index)

    if (
        sitk.GetArrayViewFromImage(label_a).sum() == 0
        or sitk.GetArrayViewFromImage(label_b).sum() == 0
    ):
        return np.nan

    hausdorff_distance = sitk.HausdorffDistanceImageFilter()
    hausdorff_distance.Execute(label_a, label_b)
    hausdorff_distance_value = hausdorff_distance.GetHausdorffDistance()

    return hausdorff_distance_value


def compute_apl(label_ref, label_test, distance_threshold_mm=3):
    """
    helper function for computing the added path length

    Args:
        label_ref (sitk.Image): The reference (ground-truth) label
        label_test (sitk.Image): The test label
        distance_threshold_mm (float): Distances under this threshold will not contribute to the
            added path length

    Returns:
        float: The total (slice-wise) added path length in mm
    """
    added_path_length_list = []
    n_slices = label_ref.GetSize()[2]

    # convert the distance threshold to voxel units
    distance = int(np.ceil(distance_threshold_mm / np.mean(label_ref.GetSpacing()[:2])))

    # iterate over each slice
    for i in range(n_slices):

        if (
            sitk.GetArrayViewFromImage(label_ref)[i].sum()
            + sitk.GetArrayViewFromImage(label_test)[i].sum()
        ) == 0:
            continue

        label_ref_contour = sitk.LabelContour(label_ref[:, :, i])
        label_test_contour = sitk.LabelContour(label_test[:, :, i])

        if distance_threshold_mm > 0:
            kernel = [int(distance) for k in range(2)]
            label_test_contour = sitk.BinaryDilate(label_test_contour, kernel)

        # mask out the locations in agreement
        added_path = sitk.MaskNegated(label_ref_contour, label_test_contour)

        # add up the voxels on the added path
        added_path_length = sitk.GetArrayViewFromImage(added_path).sum()
        added_path_length_list.append(added_path_length)

    return added_path_length_list


def compute_metric_total_apl(label_ref, label_test, distance_threshold_mm=3):
    """Compute the total (slice-wise) added path length in mm

    This operates on transaxial slices, which are assumed to be in the z-dimension (axis=2).

    Args:
        label_ref (sitk.Image): The reference (ground-truth) label
        label_test (sitk.Image): The test label
        distance_threshold_mm (float): Distances under this threshold will not contribute to the
            added path length

    Returns:
        float: The total (slice-wise) added path length in mm
    """

    added_path_length_list = compute_apl(
        label_ref, label_test, distance_threshold_mm=distance_threshold_mm
    )

    return np.sum(added_path_length_list) * np.mean(label_ref.GetSpacing()[:2])


def compute_metric_mean_apl(label_ref, label_test, distance_threshold_mm=3):
    """Compute the mean (slice-wise) added path length in mm

    This operates on transaxial slices, which are assumed to be in the z-dimension (axis=2).

    Args:
        label_ref (sitk.Image): The reference (ground-truth) label
        label_test (sitk.Image): The test label
        distance_threshold_mm (float): Distances under this threshold will not contribute to the
            added path length

    Returns:
        float: The mean (slice-wise) added path length in mm
    """

    added_path_length_list = compute_apl(
        label_ref, label_test, distance_threshold_mm=distance_threshold_mm
    )

    return np.mean(added_path_length_list) * np.mean(label_ref.GetSpacing()[:2])



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from pathlib import Path

import pydicom
import numpy as np
import SimpleITK as sitk

from skimage.draw import polygon


logger = logging.getLogger(__name__)


def read_dicom_image(dicom_path):
    """Read a DICOM image series

    Args:
        dicom_path (str|pathlib.Path): Path to the DICOM series to read

    Returns:
        sitk.Image: The image as a SimpleITK Image
    """
    dicom_images = sitk.ImageSeriesReader().GetGDCMSeriesFileNames(str(dicom_path))
    return sitk.ReadImage(dicom_images)


def read_dicom_struct_file(filename):
    """Read a DICOM RTSTRUCT file

    Args:
        filename (str|pathlib.Path): Path to the RTSTRUCT to read

    Returns:
        pydicom.Dataset: The RTSTRUCT as a DICOM Dataset
    """
    dicom_struct_file = pydicom.read_file(filename, force=True)
    return dicom_struct_file


def fix_missing_data(contour_data):
    """Fixed a set of contour data if there are values missing

    Args:
        contour_data (pydicom.Sequence): The contour sequence from the DICOM object

    Returns:
        np.array: The array of contour data with missing values fixed
    """
    contour_data = np.array(contour_data)
    if contour_data.any() == "":
        logger.debug("Missing values detected.")
        missing_values = np.where(contour_data == "")[0]
        if missing_values.shape[0] > 1:
            logger.debug("More than one value missing, fixing this isn't implemented yet...")
        else:
            logger.debug("Only one value missing.")
            missing_index = missing_values[0]
            missing_axis = missing_index % 3
            if missing_axis == 0:
                logger.debug("Missing value in x axis: interpolating.")
                if missing_index > len(contour_data) - 3:
                    lower_value = contour_data[missing_index - 3]
                    upper_value = contour_data[0]
                elif missing_index == 0:
                    lower_value = contour_data[-3]
                    upper_value = contour_data[3]
                else:
                    lower_value = contour_data[missing_index - 3]
                    upper_value = contour_data[missing_index + 3]
                contour_data[missing_index] = 0.5 * (lower_value + upper_value)
            elif missing_axis == 1:
                logger.debug("Missing value in y axis: interpolating.")
                if missing_index > len(contour_data) - 2:
                    lower_value = contour_data[missing_index - 3]
                    upper_value = contour_data[1]
                elif missing_index == 0:
                    lower_value = contour_data[-2]
                    upper_value = contour_data[4]
                else:
                    lower_value = contour_data[missing_index - 3]
                    upper_value = contour_data[missing_index + 3]
                contour_data[missing_index] = 0.5 * (lower_value + upper_value)
            else:
                logger.debug("Missing value in z axis: taking slice value")
                temp = contour_data[2::3].tolist()
                temp.remove("")
                contour_data[missing_index] = np.min(np.array(temp, dtype=np.double))
    return contour_data


def transform_point_set_from_dicom_struct(dicom_image, dicom_struct, spacing_override=None):
    """Converts a set of points from a DICOM RTSTRUCT into a mask array

    Args:
        dicom_image (sitk.Image): The reference image
        dicom_struct (pydicom.Dataset): The DICOM RTSTRUCT
        spacing_override (list): The spacing to override. Defaults to None

    Returns:
        tuple: Returns a list of masks and a list of structure names
    """

    if spacing_override:
        current_spacing = list(dicom_image.GetSpacing())
        new_spacing = tuple(
            [
                current_spacing[k] if spacing_override[k] == 0 else spacing_override[k]
                for k in range(3)
            ]
        )
        dicom_image.SetSpacing(new_spacing)

    struct_point_sequence = {cs.ReferencedROINumber: cs for cs in dicom_struct.ROIContourSequence}

    struct_list = []
    final_struct_name_sequence = []

    for struct_ds in dicom_struct.StructureSetROISequence:
        image_blank = np.zeros(dicom_image.GetSize()[::-1], dtype=np.uint8)

        struct_name = "_".join(struct_ds.ROIName.split())
        struct_index = struct_ds.ROINumber
        logger.debug("Converting structure %s with name: %s", struct_index, struct_name)

        if not struct_index in struct_point_sequence:
            logger.debug("No ROIContourSequence found for this structure, skipping.")
            continue

        if not hasattr(struct_point_sequence[struct_index], "ContourSequence"):
            logger.debug("No ContourSequence found for this structure, skipping.")
            continue

        if len(struct_point_sequence[struct_index].ContourSequence) == 0:
            logger.debug("Contour sequence empty for this structure, skipping.")
            continue

        if (
            not struct_point_sequence[struct_index].ContourSequence[0].ContourGeometricType
            == "CLOSED_PLANAR"
        ):
            logger.debug("This is not a closed planar structure, skipping.")
            continue

        # Track in case something goes wrong in here we will skip the contour
        skip_contour = False
        for sl in range(len(struct_point_sequence[struct_index].ContourSequence)):

            contour_data = fix_missing_data(
                struct_point_sequence[struct_index].ContourSequence[sl].ContourData
            )

            struct_slice_contour_data = np.array(contour_data, dtype=np.double)
            vertex_arr_physical = struct_slice_contour_data.reshape(
                struct_slice_contour_data.shape[0] // 3, 3
            )

            point_arr = np.array(
                [dicom_image.TransformPhysicalPointToIndex(i) for i in vertex_arr_physical]
            ).T

            [x_vertex_arr_image, y_vertex_arr_image] = point_arr[[0, 1]]
            z_index = point_arr[2][0]
            if np.any(point_arr[2] != z_index):
                logger.debug("Error: axial slice index varies in contour. Skipping Contour.")
                logger.debug("Structure:   %s", struct_name)
                logger.debug("Slice index: %d", z_index)
                skip_contour = True
                break

            if z_index >= dicom_image.GetSize()[2]:
                logger.debug("Warning: Slice index greater than image size. Skipping slice.")
                logger.debug("Structure:   %s", struct_name)
                logger.debug("Slice index: %d", z_index)
                continue

            slice_arr = np.zeros(image_blank.shape[-2:], dtype=np.uint8)

            filled_indices_x, filled_indices_y = polygon(
                x_vertex_arr_image, y_vertex_arr_image, shape=slice_arr.shape
            )
            slice_arr[filled_indices_y, filled_indices_x] = 1

            image_blank[z_index] += slice_arr

        if not skip_contour:
            struct_image = sitk.GetImageFromArray(1 * (image_blank > 0))
            struct_image.CopyInformation(dicom_image)
            struct_list.append(sitk.Cast(struct_image, sitk.sitkUInt8))
            final_struct_name_sequence.append(struct_name)

    return struct_list, final_struct_name_sequence


def convert_rtstruct(
    dcm_img,
    dcm_rt_file,
    prefix="Struct_",
    output_dir=".",
    output_img=None,
    spacing=None,
    replace_slashes_with="",
):
    """Convert a DICOM RTSTRUCT to NIFTI masks.

    The masks are stored as NIFTI files in the output directory

    Args:
        dcm_img (str|pathlib.Path): Path to the reference DICOM image series
        dcm_rt_file (str|pathlib.Path): Path to the DICOM RTSTRUCT file
        prefix (str, optional): The prefix to give the output files. Defaults to "Struct_".
        output_dir (str|pathlib.Path, optional): Path to the output directory. Defaults to ".".
        output_img (str|pathlib.Path, optional): If set, write the reference image to this file as
                                                 in NIFTI format. Defaults to None.
        spacing (list, optional): Values of image spacing to override. Defaults to None.
        replace_slashes_with (str, optional): String to replace "/" and "\" with. Set to None
            disable replacement of slashes. Defaults to "".
    """

    logger.debug("Converting RTStruct: %s", dcm_rt_file)
    logger.debug("Using image series: %s", dcm_img)
    logger.debug("Output file prefix: %s", prefix)
    logger.debug("Output directory: %s", output_dir)

    dicom_image = read_dicom_image(dcm_img)
    dicom_struct = read_dicom_struct_file(dcm_rt_file)

    if not isinstance(output_dir, Path):
        output_dir = Path(output_dir)

    if output_dir.exists():
        output_dir.mkdir(exist_ok=True, parents=True)

    image_output_path = None
    if output_img is not None:

        if not isinstance(output_img, Path):
            if not output_img.endswith(".nii.gz"):
                output_img = f"{output_img}.nii.gz"
            output_img = output_dir.joinpath(output_img)

        image_output_path = output_img

        logger.debug("Image series to be converted to: %s", image_output_path)

    if spacing:

        if isinstance(spacing, str):
            spacing = [float(i) for i in spacing.split(",")]

        logger.debug("Overriding image spacing with: %s", spacing)

    struct_list, struct_name_sequence = transform_point_set_from_dicom_struct(
        dicom_image, dicom_struct, spacing
    )
    logger.debug("Converted all structures. Writing output.")
    for struct_index, struct_image in enumerate(struct_list):
        struct_name = struct_name_sequence[struct_index]

        if replace_slashes_with is not None:
            struct_name = struct_name.replace("/", replace_slashes_with)
            struct_name = struct_name.replace("\\", replace_slashes_with)

        out_name = f"{prefix}{struct_name}.nii.gz"
        out_name = output_dir.joinpath(out_name)

        logger.debug("Writing file to: %s", output_dir)
        sitk.WriteImage(struct_image, str(out_name))

    if image_output_path is not None:
        sitk.WriteImage(dicom_image, str(image_output_path))



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import SimpleITK as sitk

import logging
logger = logging.getLogger(__name__)


def registration_command_iteration(method):
    """
    Utility function to print information during (rigid, similarity, translation, B-splines)
    registration
    """
    print("{0:3} = {1:10.5f}".format(method.GetOptimizerIteration(), method.GetMetricValue()))


def stage_iteration(method):
    """
    Utility function to print information during stage change in registration
    """
    print(f"Number of parameters = {method.GetInitialTransform().GetNumberOfParameters()}")


def deformable_registration_command_iteration(method):
    """
    Utility function to print information during demons registration
    """
    print("{0:3} = {1:10.5f}".format(method.GetElapsedIterations(), method.GetMetric()))


def control_point_spacing_distance_to_number(image, grid_spacing):
    """
    Convert grid spacing specified in distance to number of control points
    """
    image_spacing = np.array(image.GetSpacing())
    image_size = np.array(image.GetSize())
    number_points = image_size * image_spacing / np.array(grid_spacing)
    return (number_points + 0.5).astype(int)


def apply_linear_transform(
    input_image,
    reference_image,
    transform,
    is_structure=False,
    default_value=0,
    interpolator=sitk.sitkNearestNeighbor,
):
    """
    Helper function for applying linear transforms.

    Args
        input_image (SimpleITK.Image): The image, to which the transform is applied
        reference_image (SimpleITK.Image): The image will be resampled into this reference space.
        transform (SimpleITK.Transform): The transformation
        is_structure (bool): Will set appropriate parameters for transforming labels/structures.
        default_value: Default (background) value. Defaults to 0.
        interpolator (int, optional): The interpolation order.
                                Available options:
                                    - SimpleITK.sitkNearestNeighbor
                                    - SimpleITK.sitkLinear
                                    - SimpleITK.sitkBSpline
                                Defaults to SimpleITK.sitkNearestNeighbor

    Returns
        (SimpleITK.Image): the transformed image
    """
    if is_structure:

        if default_value != 0 or interpolator != sitk.sitkNearestNeighbor:
            logger.warning(
                "is_structure is set to True, but you have set default_value "
                "and/or interpolator. default_value and/or interpolator will be overwritten."
            )

        default_value = 0
        interpolator = sitk.sitkNearestNeighbor

    return apply_transform(
        input_image=input_image,
        reference_image=reference_image,
        transform=transform,
        default_value=default_value,
        interpolator=interpolator,
    )


def apply_deformable_transform(
    input_image,
    transform,
    is_structure=False,
    default_value=0,
    interpolator=sitk.sitkNearestNeighbor,
):
    """
    Helper function for applying deformable transforms.

    Args
        input_image (SimpleITK.Image): The image, to which the transform is applied
        reference_image (SimpleITK.Image): The image will be resampled into this reference space.
        transform (SimpleITK.Transform): The transformation
        is_structure (bool): Will set appropriate parameters for transforming labels/structures.
        default_value: Default (background) value. Defaults to 0.
        interpolator (int, optional): The interpolation order.
                                Available options:
                                    - SimpleITK.sitkNearestNeighbor
                                    - SimpleITK.sitkLinear
                                    - SimpleITK.sitkBSpline
                                Defaults to SimpleITK.sitkNearestNeighbor

    Returns
        (SimpleITK.Image): the transformed image
    """

    if is_structure:

        if default_value != 0 or interpolator != sitk.sitkNearestNeighbor:
            logger.warning(
                "is_structure is set to True, but you have set default_value "
                "and/or interpolator. default_value and/or interpolator will be overwritten."
            )

        default_value = 0
        interpolator = sitk.sitkNearestNeighbor

    return apply_transform(
        input_image=input_image,
        reference_image=None,
        transform=transform,
        default_value=default_value,
        interpolator=interpolator,
    )


def apply_transform(
    input_image,
    reference_image=None,
    transform=None,
    default_value=0,
    interpolator=sitk.sitkNearestNeighbor,
):
    """
    Transform a volume of structure with the given deformation field.

    Args
        input_image (SimpleITK.Image): The image, to which the transform is applied
        reference_image (SimpleITK.Image): The image will be resampled into this reference space.
        transform (SimpleITK.Transform): The transformation
        default_value: Default (background) value. Defaults to 0.
        interpolator (int, optional): The interpolation order.
                                Available options:
                                    - SimpleITK.sitkNearestNeighbor
                                    - SimpleITK.sitkLinear
                                    - SimpleITK.sitkBSpline
                                Defaults to SimpleITK.sitkNearestNeighbor

    Returns
        (SimpleITK.Image): the transformed image

    """
    original_image_type = input_image.GetPixelID()

    resampler = sitk.ResampleImageFilter()

    if reference_image:
        resampler.SetReferenceImage(reference_image)
    else:
        resampler.SetReferenceImage(input_image)

    if transform:
        resampler.SetTransform(transform)

    resampler.SetDefaultPixelValue(default_value)
    resampler.SetInterpolator(interpolator)

    output_image = resampler.Execute(input_image)
    output_image = sitk.Cast(output_image, original_image_type)

    return output_image


def smooth_and_resample(
    image,
    isotropic_voxel_size_mm=None,
    shrink_factor=None,
    smoothing_sigma=None,
    interpolator=sitk.sitkLinear,
):
    """
    Args:
        image (SimpleITK.Image): The image we want to resample.
        isotropic_voxel_size_mm (float | None): New voxel size in millimetres
        shrink_factor (list | float): A number greater than one, such that the new image's size is
            original_size/shrink_factor. Can also be specified independently for each
            dimension (sagittal, coronal, axial).
        smoothing_sigma (list | float): Scale for Gaussian smoothing, this is in physical
            (image spacing) units, not pixels. Can also be specified independently for
            each dimension (sagittal, coronal, axial).
    Return:
        SimpleITK.Image: Image which is a result of smoothing the input and then resampling
        it using the specified Gaussian kernel and shrink factor.
    """
    if smoothing_sigma:
        if hasattr(smoothing_sigma, "__iter__"):
            smoothing_variance = [i * i for i in smoothing_sigma]
        else:
            smoothing_variance = (smoothing_sigma ** 2,) * 3

        maximum_kernel_width = int(
            max([8 * j * i for i, j in zip(image.GetSpacing(), smoothing_variance)])
        )

        image = sitk.DiscreteGaussian(image, smoothing_variance, maximum_kernel_width)

    original_spacing = image.GetSpacing()
    original_size = image.GetSize()

    if shrink_factor and isotropic_voxel_size_mm:
        raise AttributeError(
            "Function must be called with either isotropic_voxel_size_mm or "
            "shrink_factor, not both."
        )

    elif isotropic_voxel_size_mm:
        scale_factor = (
            isotropic_voxel_size_mm * np.ones_like(image.GetSize()) / np.array(image.GetSpacing())
        )
        new_size = [int(sz / float(sf) + 0.5) for sz, sf in zip(original_size, scale_factor)]

    elif shrink_factor:
        if isinstance(shrink_factor, list):
            new_size = [int(sz / float(sf) + 0.5) for sz, sf in zip(original_size, shrink_factor)]
        else:
            new_size = [int(sz / float(shrink_factor) + 0.5) for sz in original_size]

    else:
        return image

    new_spacing = [
        ((size_o_i - 1) * spacing_o_i) / (size_n_i - 1)
        for size_o_i, spacing_o_i, size_n_i in zip(original_size, original_spacing, new_size)
    ]

    return sitk.Resample(
        image,
        new_size,
        sitk.Transform(),
        interpolator,
        image.GetOrigin(),
        new_spacing,
        image.GetDirection(),
        0.0,
        image.GetPixelID(),
    )


def convert_mask_to_distance_map(mask, squared_distance=False, normalise=False):
    """
    Generate a distance map from a binary label.

    Args:
        mask ([SimpleITK.Image]): A binary label.
        squared_distance (bool, optional): Option to use the squared distance. Defaults to False.
        normalise (bool, optional): Normalise output to [0,1]. Defaults to False.

    Returns:
        [SimpleITK.Image]: The distance map as an image.
    """
    arr = sitk.GetArrayViewFromImage(mask)
    vals = np.unique(arr[arr > 0])
    if len(vals) > 2:
        # There is more than one value! We need to threshold at the median
        cutoff = np.median(vals)
        mask = sitk.BinaryThreshold(mask, cutoff, np.max(vals).astype(float))

    raw_map = sitk.SignedMaurerDistanceMap(
        mask,
        insideIsPositive=True,
        squaredDistance=squared_distance,
        useImageSpacing=True,
    )

    if normalise:
        return raw_map / (sitk.GetArrayFromImage(raw_map).max())
    else:
        return raw_map


def convert_mask_to_reg_structure(mask, expansion=(0, 0, 0), scale=lambda x: x):
    """
    Generate a mask-like image to make structure-guided registration more
    realistic via internal deformation within a binary mask.

    Args:
        mask ([SimpleITK.Image]): The binary label.
        expansion (int, optional): For improved smoothness on the surface
                                   (particularly complex structures) it will
                                   often help to use some binary dilation. This
                                   parameter defines the expansion in mm.
                                   Defaults to 1.
        scale ([function], optional): Defines scaling to the distance map.
                                      For example: lambda x:sitk.Log(x).
                                      Defaults to lambda x:x.

    Returns:
        [SimpleITK.Image]: [description]
    """
    arr = sitk.GetArrayViewFromImage(mask)
    vals = np.unique(arr[arr > 0])
    if len(vals) > 2:
        # There is more than one value! We need to threshold at the median
        cutoff = np.median(vals)
        mask = sitk.BinaryThreshold(mask, cutoff, np.max(vals).astype(float))

    if not hasattr(expansion, "__iter__"):
        expansion = [int(expansion / i) for i in mask.GetSpacing()]
    if any(expansion):
        mask = sitk.BinaryDilate(mask, expansion)

    distance_map = sitk.Cast(
        convert_mask_to_distance_map(mask, squared_distance=False), sitk.sitkFloat64
    )

    distance_map = sitk.Mask(
        distance_map,
        mask,
    )

    scaled_distance_map = distance_map / (sitk.GetArrayViewFromImage(distance_map).max())

    return scale(scaled_distance_map)



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)
import psutil

from flask import Flask, request, render_template, jsonify

from platipy.backend import app, celery, log_file_path
from platipy.dicom.communication import DicomConnector
from .models import db, APIKey


@app.route("/endpoint/add", methods=["GET"])
def add_endpoint():

    return render_template("endpoint_add.html", data=app.data)


@app.route("/log", methods=["GET"])
def fetch_log():

    log = []
    with open(log_file_path) as f:

        for line in f:
            log.append(line.replace("\n", ""))

    return jsonify({"log": log})


@app.route("/endpoint/<id>", methods=["GET", "POST"])
def view_endpoint(id):

    endpoint = None
    for e in app.data["endpoints"]:
        if e["id"] == int(id):
            endpoint = e

    status = ""
    # Check if the last is still running
    if endpoint["endpointType"] == "listener":
        if "task_id" in endpoint:
            task = listen_task.AsyncResult(endpoint["task_id"])
            status = task.info.get("status", "")
            if "Error" in status:
                kill_task(endpoint["task_id"])

    return render_template(
        "endpoint_view.html",
        data=app.data,
        endpoint=endpoint,
        status=status,
        format_settings=lambda x: json.dumps(x, indent=4),
    )


@app.route("/status", methods=["GET"])
def fetch_status():

    celery_running = False
    if celery.control.inspect().active():
        celery_running = True
    status_context = {"celery": celery_running}
    status_context["algorithms"] = []
    for a in app.algorithms:
        algorithm = app.algorithms[a]
        status_context["algorithms"].append(
            {"name": algorithm.name, "default_settings": algorithm.default_settings}
        )

    dicom_connector = DicomConnector(
        port=app.dicom_listener_port, ae_title=app.dicom_listener_aetitle
    )
    dicom_listening = False
    if dicom_connector.verify():
        dicom_listening = True
    status_context["dicom_listener"] = {
        "port": app.dicom_listener_port,
        "aetitle": app.dicom_listener_aetitle,
        "listening": dicom_listening,
    }

    status_context["ram_usage"] = psutil.virtual_memory()._asdict()
    status_context["disk_usage"] = psutil.disk_usage("/")._asdict()
    status_context["cpu_usage"] = psutil.cpu_percent()

    status_context["applications"] = []
    for ak in APIKey.query.all():
        # Hide Key from the dashboard
        status_context["applications"].append({"name": ak.name, "key": "XXX"})

    return jsonify(status_context)


@app.route("/")
def dashboard():
    """Entry point to the dashboard of the application"""

    return render_template("dashboard.html", data={})



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import logging

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

import numpy as np
import SimpleITK as sitk

from platipy.imaging.utils.crop import label_to_roi
from platipy.imaging.visualisation.utils import (
    VisualiseBoundingBox,
    VisualiseContour,
    VisualiseScalarOverlay,
    VisualiseVectorOverlay,
    VisualiseComparisonOverlay,
    return_slice,
    subsample_vector_field,
    vector_image_grid,
    reorientate_vector_field,
    generate_comparison_colormix,
    project_onto_arbitrary_plane,
)

logger = logging.getLogger(__name__)

"""
This Python script comprises two contributions to the code base:
1) A bunch of helpful visualisation "helper" functions

2) A visualisation class used to generate figures of images, contours, vector fields, and more!
"""


class ImageVisualiser:
    """Class to assist with visualising images and overlaying contours, scalars and bounding
    boxes."""

    def __init__(
        self,
        image,
        cut=None,
        axis="ortho",
        window=None,
        figure_size_in=10,
        limits=None,
        colormap=plt.cm.get_cmap("Greys_r"),
        origin="normal",
        projection=False,
    ):
        self.__set_image(image)
        self.__contours = []
        self.__bounding_boxes = []
        self.__scalar_overlays = []
        self.__vector_overlays = []
        self.__comparison_overlays = []
        self.__show_legend = False
        self.__show_colorbar = False
        self.__figure = None
        self.__figure_size = figure_size_in
        self.__window = window
        self.__axis = axis
        self.__cut = cut
        self.__limits = limits
        self.__colormap = colormap
        self.__origin = origin
        self.__projection = projection
        self.__image_view = None
        self.__scalar_view = None
        self.__contour_colormap = None

        self.clear()

    def __set_image(self, image):
        self.__image = image

    def __set_labelmap(self, labelmap, labels=None):

        # TODO: Convert label map to binary masks for display

        raise NotImplementedError

    image = property(fset=__set_image)
    # contours = property(fset=__set_contours)
    labelmap = property(fset=__set_labelmap)

    def clear(self):
        """Clear all overlays"""

        self.__contours = []
        self.__bounding_boxes = []
        self.__scalar_overlays = []
        self.__comparison_overlays = []
        self.__vector_overlays = []

    def set_limits_from_label(self, label, expansion=2):
        """Sets the limits of the axes to the bounds of the given label.

        Args:
            label (sitk.Image): The label around which to set the limits
            expansion (list | float, optional): Expansion (in mm) around the label.
                Defaults to 2.
        """

        (sag_size, cor_size, ax_size), (sag_0, cor_0, ax_0) = label_to_roi(
            label, expansion_mm=expansion
        )

        if self.__axis == "ortho":
            self.__limits = [
                ax_0,
                ax_0 + ax_size,
                cor_0,
                cor_0 + cor_size,
                sag_0,
                sag_0 + sag_size,
            ]

        if self.__axis == "x":
            self.__limits = [cor_0, cor_0 + cor_size, ax_0, ax_0 + ax_size]
        if self.__axis == "y":
            self.__limits = [sag_0, sag_0 + sag_size, ax_0, ax_0 + ax_size]
        if self.__axis == "z":
            self.__limits = [sag_0, sag_0 + sag_size, cor_0, cor_0 + cor_size]

    def add_contour(
        self,
        contour,
        name=None,
        color=None,
        colormap=plt.cm.get_cmap("rainbow"),
        linewidth=2,
        linestyle="solid",
        show_legend=True,
    ):
        """Add a contour as overlay

        Args:
            contour (sitk.Image|dict): Contour mask or dict containing contour masks.
            name (str, optional): Name to give the contour (only used if passing sitk.Image as
                                  contour). Defaults to None.
            color (str|tuple|list, optional): The color to use when drawing the contour(s).
                                              Defaults to None.

        Raises:
            ValueError: Contour must be dict of sitk.Image.
            ValueError: If passing a dict for contour, all values must be sitk.Image.
        """

        self.__show_legend = show_legend

        if isinstance(contour, dict):

            if not all(map(lambda i: isinstance(i, sitk.Image), contour.values())):
                raise ValueError("When passing dict, all values must be of type SimpleITK.Image")

            for contour_name in contour:

                if isinstance(color, dict):
                    try:
                        contour_color = color[contour_name]
                    except AttributeError:
                        contour_color = None
                else:
                    contour_color = color

                visualise_contour = VisualiseContour(
                    contour[contour_name],
                    contour_name,
                    color=contour_color,
                    linewidth=linewidth,
                    linestyle=linestyle,
                )
                self.__contours.append(visualise_contour)

        elif isinstance(contour, sitk.Image):

            # Use a default name if not specified
            if name is None:
                name = "contour"
                self.__show_legend = False

            visualise_contour = VisualiseContour(
                contour, name, color=color, linewidth=linewidth, linestyle=linestyle
            )
            self.__contours.append(visualise_contour)
        else:

            raise ValueError(
                "Contours should be represented as a dict with contour name as key "
                "and sitk.Image as value, or as an sitk.Image and passing the contour_name"
            )

        self.__contour_colormap = colormap

    def add_scalar_overlay(
        self,
        scalar_image,
        name=None,
        colormap=plt.cm.get_cmap("cubehelix"),
        alpha=0.75,
        min_value=False,
        max_value=False,
        discrete_levels=False,
        mid_ticks=False,
        show_colorbar=True,
        norm=None,
        projection=False,
    ):
        """Overlay a scalar image on to the existing image

        Args:
            scalar_image sitk.Image|dict): Scalar image or dict containing scalar images.
            name (str, optional): Name to give the scalar image (only used if passing sitk.Image as
                                  scalar image). Defaults to None.
            colormap (matplotlib.colors.Colormap, optional): The colormap to be used when
                                                             overlaying this scalar image. Defaults
                                                             to plt.cm.get_cmap("cubehelix").
            alpha (float, optional): Alpha to apply to overlay. Defaults to 0.75.
            min_value (float, optional): Values below this value aren't rendered. Defaults to 0.1.

        Raises:
            ValueError: Scalar overlay must be dict of sitk.Image.
            ValueError: If passing a dict for contour, all values must be sitk.Image.
        """

        self.__show_colorbar = True

        if isinstance(scalar_image, dict):

            if not all(map(lambda i: isinstance(i, sitk.Image), scalar_image.values())):
                raise ValueError("When passing dict, all values must be of type SimpleITK.Image")

            for name in scalar_image:
                visualise_scalar = VisualiseScalarOverlay(
                    scalar_image[name],
                    name,
                    colormap=colormap,
                    alpha=alpha,
                    min_value=min_value,
                    max_value=max_value,
                    discrete_levels=discrete_levels,
                    mid_ticks=mid_ticks,
                    show_colorbar=show_colorbar,
                    norm=norm,
                    projection=projection,
                )
                self.__scalar_overlays.append(visualise_scalar)

        elif isinstance(scalar_image, sitk.Image):

            # Use a default name if not specified
            if not name:
                name = "Value"
                self.__show_legend = False

            visualise_scalar = VisualiseScalarOverlay(
                scalar_image,
                name,
                colormap=colormap,
                alpha=alpha,
                min_value=min_value,
                max_value=max_value,
                discrete_levels=discrete_levels,
                mid_ticks=mid_ticks,
                show_colorbar=show_colorbar,
                norm=norm,
                projection=projection,
            )
            self.__scalar_overlays.append(visualise_scalar)
        else:

            raise ValueError(
                "Contours should be represented as a dict with contour name as key "
                "and sitk.Image as value, or as an sitk.Image and passing the contour_name"
            )

    def add_vector_overlay(
        self,
        vector_image,
        min_value=False,
        max_value=False,
        colormap=plt.cm.get_cmap("inferno"),
        discrete_levels=False,
        mid_ticks=False,
        alpha=0.75,
        arrow_scale=1,
        arrow_width=1,
        subsample=8,
        color_function="magnitude",
        show_colorbar=True,
        name=None,
    ):
        """Overlay a vector field on to the existing image

        Args:
            vector_image sitk.Image|dict): Vector image (will be displayed as ).
            name (str, optional): Name to give the vector field (only used if passing
                                  sitk.Image as vector field). Defaults to None.
            colormap (matplotlib.colors.Colormap, optional): The colormap to be used when
                                                             overlaying this vector field. Defaults
                                                             to plt.cm.get_cmap("inferno").
            alpha (float, optional): Alpha to apply to overlay vectors. Defaults to 0.75.
            arrow_scale (float, optional): Relative scaling of vectors. Defaults to 1.
            arrow_width (float, optional): Width of vector field arrow. Defaults to 1.
            subsample (int, optional): Defines to subsampling ratio of displayed vectors.
                                       Defaults to 8.
            color_function (str, optional): Determines how vectors are colored. Options:
                                            'perpendicular' - vectors colored by perpendicular
                                            value
                                            'magnitude' - vectors colored by magnitude.

        Raises:
            ValueError: Vector overlay must be of type sitk.Image.
        """

        if (
            isinstance(vector_image, sitk.Image)
            and vector_image.GetNumberOfComponentsPerPixel() > 1
        ):

            # Use a default name if not specified
            if not name:
                name = "Vector Field [mm]"
                self.__show_legend = False

            visualise_vector_field = VisualiseVectorOverlay(
                vector_image,
                min_value=min_value,
                max_value=max_value,
                colormap=colormap,
                discrete_levels=discrete_levels,
                mid_ticks=mid_ticks,
                alpha=alpha,
                arrow_scale=arrow_scale,
                arrow_width=arrow_width,
                subsample=subsample,
                color_function=color_function,
                show_colorbar=show_colorbar,
                name=name,
            )
            self.__vector_overlays.append(visualise_vector_field)
        else:

            raise ValueError("Vector field should be sitk.Image (of vector type).")

    def add_comparison_overlay(self, image, name=None, color_rotation=0.35):
        """Overlay a comparison image on the existing image

        Args:
            image sitk.Image): Image (will be displayed as a comparison).
            name (str, optional): Name to give the image. Defaults to None.
            color_rotation (float, optional): Defines the hue of the original image (0 - 0.5).

        Raises:
            ValueError: Comparison overlay must be of type sitk.Image.
        """

        if isinstance(image, sitk.Image):

            visualise_comparison = VisualiseComparisonOverlay(
                image, name, color_rotation=color_rotation
            )
            self.__comparison_overlays.append(visualise_comparison)
        else:

            raise ValueError("Image should be sitk.Image.")

    def add_bounding_box(self, bounding_box, name=None, color="r", linewidth=2):
        """Add a bounding box to draw.

        Args:
            bounding_box (dict|list|tuple|sitk.Image): Dictionary of bounding boxes where list key
                                    is the name of the bounding box. If list or tuple then pass
                                    bounding box in format (x, y, z, w, h, d). If sitk.Image then
                                    mask is assumed and bounding box around this mask is used.
            name (list, optional): Name of this bounding box if bounding_box is list, tuple or
                                   sitk.Image. Defaults to None.
            color (str|tuple|list, optional): The color to use when drawing the bounding box(es).
                                              Defaults to None.
            linewidth (int, optional): The line width of the bounding box(es). Defaults to 2.

        Raises:
            ValueError: Raised if input is not a list or tuple of length 6 or and sitk.Image.
        """

        self.__show_legend = True

        if isinstance(bounding_box, dict):

            if not all(
                map(
                    lambda i: isinstance(i, (list, tuple, sitk.Image)) and len(i) == 6,
                    bounding_box.values(),
                )
            ):
                raise ValueError("All values must be of type list or tuple with length 6")

            for name in bounding_box:
                visualise_bounding_box = VisualiseBoundingBox(
                    bounding_box[name], name=name, color=color, linewidth=linewidth
                )
                self.__bounding_boxes.append(visualise_bounding_box)

        elif isinstance(bounding_box, (list, tuple, sitk.Image)):

            # Use a default name if not specified
            if name is None:
                name = "Bounding box"

            visualise_bounding_box = VisualiseBoundingBox(
                bounding_box, name=name, color=color, linewidth=linewidth
            )
            self.__bounding_boxes.append(visualise_bounding_box)

        else:
            raise ValueError(
                "Bounding boxes should be represented as a dict with bounding box name as key "
                "and list or tuple as value"
            )

    def show(self, interact=False):
        """Render the image with all overlays"""
        if len(self.__comparison_overlays) == 0:
            self._display_slice()
        else:
            self._overlay_comparison()

        self._overlay_scalar_field()
        self._overlay_vector_field()
        self._overlay_contours()
        self._overlay_bounding_boxes()

        self._adjust_view()

        if interact:
            logger.warning("Interactive mode not yet implemented")
            # self.interact_adjust_slice()

        self.__figure.canvas.draw()
        self._add_legend()
        self.__figure.set_facecolor("white")

        return self.__figure

    def _display_slice(self):
        """Display the configured image slice"""

        image = self.__image
        nda = sitk.GetArrayFromImage(image)

        (ax_size, cor_size, sag_size) = nda.shape[:3]

        window = self.__window
        if window is None:
            # We will choose it ourselves!

            lower = nda.min()

            # Check if we *probably* have a CT
            if lower < -900:
                # Just set a decent CT window
                # Somewhere around soft tissue
                window = (-250, 600)

            # Otherwise just pick a reasonable upper limit
            else:
                upper = np.percentile(nda, 99)
                window = (lower, upper - lower)
        try:
            logger.info(
                "Found a (z,y,x,%s) dimensional array - assuming this is an RGB" "image.",
                nda.shape[3],
            )
            nda /= nda.max()
        except ValueError:
            logger.warning("Problem converting RGB image to np.ndarray.")
        except IndexError:
            pass

        sp_plane, _, sp_slice = image.GetSpacing()
        asp = (1.0 * sp_slice) / sp_plane

        # Get the size - this is used for extent
        size_sag, size_cor, size_ax = self.__image.GetSize()
        extent_dict = {
            "x": (0, size_cor, 0, size_ax),
            "y": (0, size_sag, 0, size_ax),
            "z": (0, size_sag, 0, size_cor),
        }

        if self.__projection is True:
            projection = "max"
        else:
            projection = self.__projection

        if self.__axis == "ortho":
            figure_size = (
                self.__figure_size,
                self.__figure_size * (asp * ax_size + cor_size) / (1.0 * sag_size + cor_size),
            )

            self.__figure, ((ax_ax, blank), (ax_cor, ax_sag)) = plt.subplots(
                2,
                2,
                figsize=figure_size,
                gridspec_kw={
                    "height_ratios": [(cor_size) / (asp * ax_size), 1],
                    "width_ratios": [sag_size, cor_size],
                },
            )
            blank.axis("off")

            if self.__cut is None:
                slice_ax = int(ax_size / 2.0)
                slice_cor = int(cor_size / 2.0)
                slice_sag = int(sag_size / 2.0)

                self.__cut = [slice_ax, slice_cor, slice_sag]

            if not self.__projection:
                s_ax = return_slice("z", self.__cut[0])
                s_cor = return_slice("y", self.__cut[1])
                s_sag = return_slice("x", self.__cut[2])

                ax_img = nda.__getitem__(s_ax)
                cor_img = nda.__getitem__(s_cor)
                sag_img = nda.__getitem__(s_sag)

            else:
                ax_img_proj = project_onto_arbitrary_plane(
                    image,
                    projection_axis=2,
                    projection_name=projection,
                    default_value=int(nda.min()),
                )
                ax_img = sitk.GetArrayFromImage(ax_img_proj)
                # ax_img = (ax_img - ax_img.min()) / (ax_img.max() - ax_img.min())

                cor_img_proj = project_onto_arbitrary_plane(
                    image,
                    projection_axis=1,
                    projection_name=projection,
                    default_value=int(nda.min()),
                )
                cor_img = sitk.GetArrayFromImage(cor_img_proj)
                # cor_img = (cor_img - cor_img.min()) / (cor_img.max() - cor_img.min())

                sag_img_proj = project_onto_arbitrary_plane(
                    image,
                    projection_axis=0,
                    projection_name=projection,
                    default_value=int(nda.min()),
                )
                sag_img = sitk.GetArrayFromImage(sag_img_proj)
                # sag_img = (sag_img - sag_img.min()) / (sag_img.max() - sag_img.min())

            ax_view = ax_ax.imshow(
                ax_img,
                aspect=1.0,
                interpolation="none",
                origin={"normal": "upper", "reversed": "lower"}[self.__origin],
                cmap=self.__colormap,
                clim=(window[0], window[0] + window[1]),
                extent=extent_dict["z"],
            )
            cor_view = ax_cor.imshow(
                cor_img,
                origin="lower",
                aspect=asp,
                interpolation="none",
                cmap=self.__colormap,
                clim=(window[0], window[0] + window[1]),
                extent=extent_dict["y"],
            )
            sag_view = ax_sag.imshow(
                sag_img,
                origin="lower",
                aspect=asp,
                interpolation="none",
                cmap=self.__colormap,
                clim=(window[0], window[0] + window[1]),
                extent=extent_dict["x"],
            )

            ax_ax.axis("off")
            ax_cor.axis("off")
            ax_sag.axis("off")

            self.__figure.subplots_adjust(
                left=0, right=1, wspace=0.01, hspace=0.01, top=1, bottom=0
            )

            self.__image_view = {
                "ax_view": ax_view,
                "cor_view": cor_view,
                "sag_view": sag_view,
            }

        else:

            if hasattr(self.__cut, "__iter__"):
                warnings.warn(
                    "You have selected a single axis and multiple slice locations, attempting to "
                    "match."
                )
                self.__cut = self.__cut[{"x": 2, "y": 1, "z": 0}[self.__axis]]

            if self.__axis == "x" or self.__axis == "sag":
                axis_view_name_consistent = "sag_view"
                figure_size = (
                    self.__figure_size,
                    self.__figure_size * (asp * ax_size) / (1.0 * cor_size),
                )
                self.__figure, ax = plt.subplots(1, 1, figsize=(figure_size))
                org = "lower"
                if not self.__cut:
                    self.__cut = int(sag_size / 2.0)

            if self.__axis == "y" or self.__axis == "cor":
                axis_view_name_consistent = "cor_view"
                figure_size = (
                    self.__figure_size,
                    self.__figure_size * (asp * ax_size) / (1.0 * sag_size),
                )
                self.__figure, ax = plt.subplots(1, 1, figsize=(figure_size))
                org = "lower"
                if not self.__cut:
                    self.__cut = int(cor_size / 2.0)

            if self.__axis == "z" or self.__axis == "ax":
                axis_view_name_consistent = "ax_view"
                asp = 1
                figure_size = (
                    self.__figure_size,
                    self.__figure_size * (asp * cor_size) / (1.0 * sag_size),
                )
                self.__figure, ax = plt.subplots(1, 1, figsize=(figure_size))
                org = {"normal": "upper", "reversed": "lower"}[self.__origin]
                if not self.__cut:
                    self.__cut = int(ax_size / 2.0)

            if not self.__projection:
                s = return_slice(self.__axis, self.__cut)
                disp_img = nda.__getitem__(s)
            else:
                disp_img_proj = project_onto_arbitrary_plane(
                    image,
                    projection_axis={"x": 0, "y": 1, "z": 2}[self.__axis],
                    projection_name=projection,
                    default_value=int(nda.min()),
                )
                disp_img = sitk.GetArrayFromImage(disp_img_proj)
                # disp_img = (disp_img - disp_img.min()) / (disp_img.max() - disp_img.min())

            s = return_slice(self.__axis, self.__cut)
            ax_indiv = ax.imshow(
                disp_img,
                aspect=asp,
                interpolation="none",
                origin=org,
                cmap=self.__colormap,
                clim=(window[0], window[0] + window[1]),
                extent=extent_dict[self.__axis],
            )
            ax.axis("off")

            self.__figure.subplots_adjust(left=0, right=1, bottom=0, top=1)

            self.__image_view = {axis_view_name_consistent: ax_indiv}

    def _overlay_comparison(self):
        """Display an overlay comparison

        Args:
            color_rotation (float, optional): The hue used to color the original image (0 - 0.5).
        """

        if len(self.__comparison_overlays) > 1:
            raise ValueError("You can only display one comparison image.")

        else:
            comparison_overlay = self.__comparison_overlays[0]

        image_original = self.__image
        nda_original = sitk.GetArrayFromImage(image_original)

        image_new = comparison_overlay.image
        nda_new = sitk.GetArrayFromImage(image_new)
        color_rotation = comparison_overlay.color_rotation

        (ax_size, cor_size, sag_size) = nda_original.shape
        sp_plane, _, sp_slice = image_original.GetSpacing()
        asp = (1.0 * sp_slice) / sp_plane

        window = self.__window
        if window is None:
            # We will choose it ourselves!

            lower = nda_original.min()

            # Check if we *probably* have a CT
            if lower < -900:
                # Just set a decent CT window
                # Somewhere around soft tissue
                window = (-250, 600)

            # Otherwise just pick a reasonable upper limit
            else:
                upper = np.percentile(nda_original, 99)
                window = (lower, upper - lower)

        # Get the size - this is used for extent
        size_sag, size_cor, size_ax = self.__image.GetSize()
        extent_dict = {
            "x": (0, size_cor, 0, size_ax),
            "y": (0, size_sag, 0, size_ax),
            "z": (0, size_sag, 0, size_cor),
        }

        if self.__axis == "ortho":
            figure_size = (
                self.__figure_size,
                self.__figure_size * (asp * ax_size + cor_size) / (1.0 * sag_size + cor_size),
            )

            self.__figure, ((ax_ax, blank), (ax_cor, ax_sag)) = plt.subplots(
                2,
                2,
                figsize=figure_size,
                gridspec_kw={
                    "height_ratios": [(cor_size) / (asp * ax_size), 1],
                    "width_ratios": [sag_size, cor_size],
                },
            )
            blank.axis("off")

            if self.__cut is None:
                slice_ax = int(ax_size / 2.0)
                slice_cor = int(cor_size / 2.0)
                slice_sag = int(sag_size / 2.0)

                self.__cut = [slice_ax, slice_cor, slice_sag]

            s_ax = return_slice("z", self.__cut[0])
            s_cor = return_slice("y", self.__cut[1])
            s_sag = return_slice("x", self.__cut[2])

            nda_colormix = generate_comparison_colormix(
                [nda_original, nda_new],
                arr_slice=s_ax,
                window=window,
                color_rotation=color_rotation,
            )

            ax_ax.imshow(
                nda_colormix,
                aspect=1.0,
                origin={"normal": "upper", "reversed": "lower"}[self.__origin],
                interpolation="none",
                extent=extent_dict["z"],
            )

            nda_colormix = generate_comparison_colormix(
                [nda_original, nda_new],
                arr_slice=s_cor,
                window=window,
                color_rotation=color_rotation,
            )

            ax_cor.imshow(
                nda_colormix,
                origin="lower",
                aspect=asp,
                interpolation="none",
                extent=extent_dict["y"],
            )

            nda_colormix = generate_comparison_colormix(
                [nda_original, nda_new],
                arr_slice=s_sag,
                window=window,
                color_rotation=color_rotation,
            )

            ax_sag.imshow(
                nda_colormix,
                origin="lower",
                aspect=asp,
                interpolation="none",
                extent=extent_dict["x"],
            )

            ax_ax.axis("off")
            ax_cor.axis("off")
            ax_sag.axis("off")

            self.__figure.subplots_adjust(
                left=0, right=1, wspace=0.01, hspace=0.01, top=1, bottom=0
            )

        else:

            if hasattr(self.__cut, "__iter__"):
                warnings.warn(
                    "You have selected a single axis and multiple slice locations, attempting to "
                    "match."
                )
                self.__cut = self.__cut[{"x": 2, "y": 1, "z": 0}[self.__axis]]

            if self.__axis == "x" or self.__axis == "sag":
                figure_size = (
                    self.__figure_size,
                    self.__figure_size * (asp * ax_size) / (1.0 * cor_size),
                )
                self.__figure, ax = plt.subplots(1, 1, figsize=(figure_size))
                org = "lower"
                if not self.__cut:
                    self.__cut = int(sag_size / 2.0)

            if self.__axis == "y" or self.__axis == "cor":
                figure_size = (
                    self.__figure_size,
                    self.__figure_size * (asp * ax_size) / (1.0 * sag_size),
                )
                self.__figure, ax = plt.subplots(1, 1, figsize=(figure_size))
                org = "lower"
                if not self.__cut:
                    self.__cut = int(cor_size / 2.0)

            if self.__axis == "z" or self.__axis == "ax":
                asp = 1
                figure_size = (
                    self.__figure_size,
                    self.__figure_size * (asp * cor_size) / (1.0 * sag_size),
                )
                self.__figure, ax = plt.subplots(1, 1, figsize=(figure_size))
                org = "upper"
                if not self.__cut:
                    self.__cut = int(ax_size / 2.0)

            s = return_slice(self.__axis, self.__cut)

            nda_colormix = generate_comparison_colormix(
                [nda_original, nda_new], arr_slice=s, window=window, color_rotation=color_rotation
            )

            ax.imshow(
                nda_colormix,
                aspect=asp,
                interpolation="none",
                origin=org,
                extent=extent_dict[self.__axis],
            )
            ax.axis("off")

            self.__figure.subplots_adjust(left=0, right=1, bottom=0, top=1)

    def _adjust_view(self):
        """adjust_view is a helper function for modifying axis limits.
        Specify *limits* when initialising the ImageVisulaiser to use.
        Alternatively, use set_limits_from_label to specify automatically.
        """

        limits = self.__limits
        origin = self.__origin

        if limits is not None:
            if self.__axis == "ortho":
                ax_ax, ax_blank, ax_cor, ax_sag = self.__figure.axes[:4]
                cax_list = self.__figure.axes[4:]

                ax_orig_0, ax_orig_1 = ax_cor.get_ylim()
                cor_orig_0, cor_orig_1 = ax_ax.get_ylim()
                sag_orig_0, sag_orig_1 = ax_ax.get_xlim()

                ax_0, ax_1, cor_0, cor_1, sag_0, sag_1 = limits

                # Perform some corrections
                ax_0, ax_1 = sorted([ax_0, ax_1])
                cor_0, cor_1 = sorted([cor_0, cor_1])
                sag_0, sag_1 = sorted([sag_0, sag_1])

                ax_orig_0, ax_orig_1 = sorted([ax_orig_0, ax_orig_1])
                cor_orig_0, cor_orig_1 = sorted([cor_orig_0, cor_orig_1])
                sag_orig_0, sag_orig_1 = sorted([sag_orig_0, sag_orig_1])

                ax_size = ax_1 - ax_0
                cor_size = cor_1 - cor_0
                sag_size = sag_1 - sag_0

                asp = ax_cor.get_aspect()

                ratio_x = ((cor_1 - cor_0) + (sag_1 - sag_0)) / (
                    (cor_orig_1 - cor_orig_0) + (sag_orig_1 - sag_orig_0)
                )
                ratio_y = (1 / asp * (cor_1 - cor_0) + (ax_1 - ax_0)) / (
                    1 / asp * (cor_orig_1 - cor_orig_0) + (ax_orig_1 - ax_orig_0)
                )

                ax_ax.set_xlim(sag_0, sag_1)

                if origin == "reversed":
                    ax_ax.set_ylim(cor_0, cor_1)
                else:
                    ax_ax.set_ylim(cor_orig_1 - cor_1, cor_orig_1 - cor_0)

                ax_cor.set_xlim(sag_0, sag_1)
                ax_cor.set_ylim(ax_0, ax_1)

                ax_sag.set_xlim(cor_0, cor_1)
                ax_sag.set_ylim(ax_0, ax_1)

                gs = gridspec.GridSpec(
                    2,
                    2,
                    height_ratios=[(cor_size) / (asp * ax_size), 1],
                    width_ratios=[sag_size, cor_size],
                )

                ax_ax.set_position(gs[0].get_position(self.__figure))
                ax_ax.set_subplotspec(gs[0])

                ax_blank.set_position(gs[1].get_position(self.__figure))
                ax_blank.set_subplotspec(gs[1])

                ax_cor.set_position(gs[2].get_position(self.__figure))
                ax_cor.set_subplotspec(gs[2])

                ax_sag.set_position(gs[3].get_position(self.__figure))
                ax_sag.set_subplotspec(gs[3])

                fig_size_x, fig_size_y = self.__figure.get_size_inches()
                fig_size_y = fig_size_y * ratio_y / ratio_x

                ax_ax_bbox = gs[0].get_position(self.__figure)

                for cax_index, cax in enumerate(cax_list):

                    cbar_width = ax_ax_bbox.width * 0.05

                    cax.set_position(
                        (
                            ax_ax_bbox.x1 + 0.02 + (cbar_width + 0.1) * cax_index,
                            ax_ax_bbox.y0 + 0.01,
                            0.05,
                            ax_ax_bbox.height - 0.02,
                        )
                    )

                self.__figure.set_size_inches(fig_size_x, fig_size_y)

            elif self.__axis in ["x", "y", "z"]:
                ax = self.__figure.axes[0]
                x_orig_0, x_orig_1 = ax.get_xlim()
                y_orig_0, y_orig_1 = ax.get_ylim()

                x_0, x_1, y_0, y_1 = limits
                # Perform some corrections
                x_0, x_1 = sorted([x_0, x_1])
                y_0, y_1 = sorted([y_0, y_1])

                if self.__axis == "z" and self.__origin == "normal":
                    y_0, y_1 = y_1, y_0

                ratio_x = np.abs(x_1 - x_0) / np.abs(x_orig_1 - x_orig_0)
                ratio_y = np.abs(y_1 - y_0) / np.abs(y_orig_1 - y_orig_0)

                ax.set_xlim(x_0, x_1)
                ax.set_ylim(y_0, y_1)

                fig_size_x, fig_size_y = self.__figure.get_size_inches()
                fig_size_y = fig_size_y * ratio_y / ratio_x

                self.__figure.set_size_inches(fig_size_x, fig_size_y)

    def _overlay_contours(self):
        """Overlay the contours on to the current figure image"""

        if len(self.__contours) == 0:
            return

        plot_dict = {}
        color_dict = {}
        lw_dict = {}
        ls_dict = {}

        color_gen_index = 0

        for contour in self.__contours:
            contour_image_resampled = sitk.Resample(contour.image, self.__image)
            plot_dict[contour.name] = contour_image_resampled

            if contour.color is not None:
                color_dict[contour.name] = contour.color
            else:
                color_map = self.__contour_colormap(np.linspace(0, 1, len(self.__contours)))

                color_dict[contour.name] = color_map[color_gen_index % 255]
                color_gen_index += 1

            lw_dict[contour.name] = contour.linewidth
            ls_dict[contour.name] = contour.linestyle

        # Test types of axes
        axes = self.__figure.axes[:4]

        # Get the size - this is used for extent
        size_sag, size_cor, size_ax = self.__image.GetSize()
        extent_dict = {
            "x": (0, size_cor, 0, size_ax),
            "y": (0, size_sag, 0, size_ax),
            "z": (0, size_sag, 0, size_cor),
        }
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="No contour levels were found within the data range."
            )

            if self.__axis in ["x", "y", "z"]:
                ax = axes[0]
                s = return_slice(self.__axis, self.__cut)

                for c_name, mask in plot_dict.items():
                    if not self.__projection:
                        contour_disp = sitk.GetArrayFromImage(mask).__getitem__(s)

                        # Force a single pixel to 1 to display all contours
                        # even if they aren't defined on a particular slice
                        if contour_disp.sum() == 0:
                            contour_disp[0, 0] = 1

                    else:
                        contour_disp_proj = project_onto_arbitrary_plane(
                            plot_dict[c_name],
                            projection_axis={"x": 0, "y": 1, "z": 2}[self.__axis],
                            projection_name="max",
                            default_value=0,
                        )
                        contour_disp = sitk.GetArrayFromImage(contour_disp_proj)

                    if self.__axis == "z":
                        origin = {"normal": "upper", "reversed": "lower"}[self.__origin]
                    else:
                        origin = "lower"

                    try:
                        ax.contour(
                            contour_disp,
                            colors=[color_dict[c_name]],
                            levels=[0.5],
                            # alpha=0.8,
                            linewidths=lw_dict[c_name],
                            linestyles=ls_dict[c_name],
                            extent=extent_dict[self.__axis],
                            origin=origin,
                        )
                        ax.plot(
                            [0],
                            [0],
                            lw=lw_dict[c_name],
                            ls=ls_dict[c_name],
                            c=color_dict[c_name],
                            label=c_name,
                        )
                    except AttributeError:
                        pass

            elif self.__axis == "ortho":
                ax_ax, _, ax_cor, ax_sag = axes

                ax = ax_ax

                s_ax = return_slice("z", self.__cut[0])
                s_cor = return_slice("y", self.__cut[1])
                s_sag = return_slice("x", self.__cut[2])

                for _, c_name in enumerate(plot_dict.keys()):

                    if not self.__projection:

                        contour_ax = sitk.GetArrayFromImage(plot_dict[c_name]).__getitem__(s_ax)
                        contour_cor = sitk.GetArrayFromImage(plot_dict[c_name]).__getitem__(s_cor)
                        contour_sag = sitk.GetArrayFromImage(plot_dict[c_name]).__getitem__(s_sag)

                        # Force a single pixel to 1 to display all contours
                        # even if they aren't defined on a particular slice
                        if contour_ax.sum() == 0:
                            contour_ax[0, 0] = 1
                        if contour_cor.sum() == 0:
                            contour_cor[0, 0] = 1
                        if contour_sag.sum() == 0:
                            contour_sag[0, 0] = 1

                    else:
                        contour_ax_proj = project_onto_arbitrary_plane(
                            plot_dict[c_name],
                            projection_axis=2,
                            projection_name="max",
                            default_value=0,
                        )
                        contour_ax = sitk.GetArrayFromImage(contour_ax_proj)

                        contour_cor_proj = project_onto_arbitrary_plane(
                            plot_dict[c_name],
                            projection_axis=1,
                            projection_name="max",
                            default_value=0,
                        )
                        contour_cor = sitk.GetArrayFromImage(contour_cor_proj)

                        contour_sag_proj = project_onto_arbitrary_plane(
                            plot_dict[c_name],
                            projection_axis=0,
                            projection_name="max",
                            default_value=0,
                        )
                        contour_sag = sitk.GetArrayFromImage(contour_sag_proj)

                    ax_ax.contour(
                        contour_ax,
                        levels=[0],
                        linewidths=lw_dict[c_name],
                        linestyles=ls_dict[c_name],
                        colors=[color_dict[c_name]],
                        extent=extent_dict["z"],
                        origin={"normal": "upper", "reversed": "lower"}[self.__origin],
                    )
                    ax_ax.plot(
                        [0],
                        [0],
                        lw=lw_dict[c_name],
                        ls=ls_dict[c_name],
                        c=color_dict[c_name],
                        label=c_name,
                    )

                    ax_cor.contour(
                        contour_cor,
                        levels=[0.5],
                        linewidths=lw_dict[c_name],
                        linestyles=ls_dict[c_name],
                        colors=[color_dict[c_name]],
                        extent=extent_dict["y"],
                        origin="lower",
                    )
                    ax_sag.contour(
                        contour_sag,
                        levels=[0.5],
                        linewidths=lw_dict[c_name],
                        linestyles=ls_dict[c_name],
                        colors=[color_dict[c_name]],
                        extent=extent_dict["x"],
                        origin="lower",
                    )

            else:
                raise ValueError('Axis is must be one of "x","y","z","ortho".')

    def _overlay_scalar_field(self):
        """Overlay the scalar image onto the existing figure"""

        for scalar_index, scalar in enumerate(self.__scalar_overlays):

            scalar_image = scalar.image
            nda = sitk.GetArrayFromImage(scalar_image)

            alpha = scalar.alpha

            if scalar.max_value is not False:
                s_max = scalar.max_value
            else:
                s_max = nda.max()

            if scalar.min_value is not False:
                s_min = scalar.min_value
            else:
                s_min = nda.min()

            if scalar.discrete_levels:
                colormap_name = scalar.colormap.name
                colormap = plt.cm.get_cmap(colormap_name, scalar.discrete_levels)

            else:
                colormap = scalar.colormap

            if scalar.norm:
                norm = scalar.norm
            else:
                norm = None

            # Get the spacing
            sp_plane, _, sp_slice = scalar_image.GetSpacing()

            # Get the aspect ratio
            asp = (1.0 * sp_slice) / sp_plane

            # Get the size - this is used for extent
            size_sag, size_cor, size_ax = self.__image.GetSize()
            extent_dict = {
                "x": (0, size_cor, 0, size_ax),
                "y": (0, size_sag, 0, size_ax),
                "z": (0, size_sag, 0, size_cor),
            }

            # projection organisation
            if scalar.projection:
                projection = scalar.projection
            else:
                projection = self.__projection

            if self.__axis == "ortho":

                ax_ax, _, ax_cor, ax_sag = self.__figure.axes[:4]
                ax = ax_ax

                if not projection:
                    s_ax = return_slice("z", self.__cut[0])
                    s_cor = return_slice("y", self.__cut[1])
                    s_sag = return_slice("x", self.__cut[2])

                    ax_img = nda.__getitem__(s_ax)
                    cor_img = nda.__getitem__(s_cor)
                    sag_img = nda.__getitem__(s_sag)

                else:
                    ax_img_proj = project_onto_arbitrary_plane(
                        scalar_image,
                        projection_axis=2,
                        projection_name=projection,
                        default_value=int(nda.min()),
                    )
                    ax_img = sitk.GetArrayFromImage(ax_img_proj)

                    cor_img_proj = project_onto_arbitrary_plane(
                        scalar_image,
                        projection_axis=1,
                        projection_name=projection,
                        default_value=int(nda.min()),
                    )
                    cor_img = sitk.GetArrayFromImage(cor_img_proj)

                    sag_img_proj = project_onto_arbitrary_plane(
                        scalar_image,
                        projection_axis=0,
                        projection_name=projection,
                        default_value=int(nda.min()),
                    )
                    sag_img = sitk.GetArrayFromImage(sag_img_proj)

                # mask images to enforce transparency
                ax_img = np.ma.masked_less_equal(ax_img, s_min)
                cor_img = np.ma.masked_less_equal(cor_img, s_min)
                sag_img = np.ma.masked_less_equal(sag_img, s_min)

                ax_view = ax_ax.imshow(
                    ax_img,
                    interpolation="none",
                    cmap=colormap,
                    clim=(s_min, s_max),
                    aspect=1,
                    origin={"normal": "upper", "reversed": "lower"}[self.__origin],
                    vmin=s_min,
                    vmax=s_max,
                    alpha=alpha,
                    norm=norm,
                    extent=extent_dict["z"],
                )

                cor_view = ax_cor.imshow(
                    cor_img,
                    interpolation="none",
                    cmap=colormap,
                    clim=(s_min, s_max),
                    origin="lower",
                    aspect=asp,
                    vmin=s_min,
                    vmax=s_max,
                    alpha=alpha,
                    norm=norm,
                    extent=extent_dict["y"],
                )

                sag_view = ax_sag.imshow(
                    sag_img,
                    interpolation="none",
                    cmap=colormap,
                    clim=(s_min, s_max),
                    origin="lower",
                    aspect=asp,
                    vmin=s_min,
                    vmax=s_max,
                    alpha=alpha,
                    norm=norm,
                    extent=extent_dict["x"],
                )

                # this is for (work-in-progress) dynamic visualisation
                self.__scalar_view = {
                    "ax_view": ax_view,
                    "cor_view": cor_view,
                    "sag_view": sag_view,
                }

            else:

                ax = self.__figure.axes[0]

                if not projection:
                    s = return_slice(self.__axis, self.__cut)
                    disp_img = nda.__getitem__(s)
                else:
                    disp_img_proj = project_onto_arbitrary_plane(
                        scalar_image,
                        projection_axis={"x": 0, "y": 1, "z": 2}[self.__axis],
                        projection_name=projection,
                        default_value=int(nda.min()),
                    )
                    disp_img = sitk.GetArrayFromImage(disp_img_proj)
                    # disp_img = (disp_img - disp_img.min()) / (disp_img.max() - disp_img.min())

                disp_img = np.ma.masked_less_equal(disp_img, s_min)

                asp = {"x": asp, "y": asp, "z": 1}[self.__axis]

                if self.__axis == "z":
                    origin = {"normal": "upper", "reversed": "lower"}[self.__origin]
                else:
                    origin = "lower"

                s = return_slice(self.__axis, self.__cut)
                ax_view = ax.imshow(
                    disp_img,
                    interpolation="none",
                    cmap=colormap,
                    clim=(s_min, s_max),
                    origin=origin,
                    aspect=asp,
                    vmin=s_min,
                    vmax=s_max,
                    alpha=alpha,
                    norm=norm,
                    extent=extent_dict[self.__axis],
                )

            if scalar.show_colorbar:

                # divider = make_axes_locatable(ax_view)
                # cax = divider.append_axes("right", size="5%", pad=0.05)

                ax_box = ax.get_position(original=False)
                cbar_width = ax_box.width * 0.05  # 5% of axis width
                cbar_color = "black"

                if self.__axis == "ortho":
                    cax = self.__figure.add_axes(
                        (
                            ax_box.x1 + 0.02 + (cbar_width + 0.1) * scalar_index,
                            ax_box.y0 * 1.025,
                            cbar_width,
                            ax_box.height - ax_box.y0 * 0.05,
                        )
                    )

                else:
                    cax = self.__figure.add_axes(
                        (
                            ax_box.x1 - 0.02 - (cbar_width + 0.1) * (scalar_index + 1),
                            0.025,
                            cbar_width,
                            ax_box.height - ax_box.y1 * 0.05,
                        )
                    )

                    # check background values
                    if np.linalg.norm(colormap(0)[:3]) < 0.1:
                        # background is dark
                        cbar_color = "white"

                cbar = self.__figure.colorbar(ax_view, cax=cax, orientation="vertical")

                # set color
                cbar.outline.set_edgecolor(color=cbar_color)
                cbar.ax.tick_params(color=cbar_color)
                cax.tick_params(axis="x", colors=cbar_color)
                cax.tick_params(axis="y", colors=cbar_color)

                cbar.set_label(scalar.name, color=cbar_color)
                cbar.solids.set_alpha(1)

                if scalar.discrete_levels:

                    if scalar.mid_ticks:

                        delta_tick = (s_max - s_min) / scalar.discrete_levels
                        cbar.set_ticks(
                            np.linspace(
                                s_min + delta_tick / 2,
                                s_max - delta_tick / 2,
                                scalar.discrete_levels,
                            )
                        )
                        cbar.set_ticklabels(np.linspace(s_min, s_max, scalar.discrete_levels))

                    else:
                        cbar.set_ticks(
                            np.linspace(
                                s_min,
                                s_max,
                                scalar.discrete_levels + 1,
                            )
                        )

    def _overlay_vector_field(self):
        """Overlay vector field onto existing figure"""

        if self.__projection and len(self.__vector_overlays) > 0:
            raise Warning("Vector overlay is not implemented in projection mode.")

        for vector_index, vector in enumerate(self.__vector_overlays):

            image = vector.image
            colormap = vector.colormap
            alpha = vector.alpha
            arrow_scale = vector.arrow_scale
            arrow_width = vector.arrow_width
            subsample = vector.subsample
            color_function = vector.color_function
            invert_field = vector.invert_field
            alpha = vector.alpha
            min_value = vector.min_value
            max_value = vector.max_value

            inverse_vector_image = image  # sitk.InvertDisplacementField(image)
            vector_nda = sitk.GetArrayFromImage(inverse_vector_image)

            if vector.discrete_levels:
                colormap_name = vector.colormap.name
                colormap = plt.cm.get_cmap(colormap_name, vector.discrete_levels)

            # Test types of axes
            axes = self.__figure.axes
            if len(axes[:4]) < 4:
                ax = axes[0]

                if hasattr(subsample, "__iter__"):
                    raise ValueError(
                        "You have selected an iterable subsampling factor for a\
                                      single axis. Behaviour undefined in this situation."
                    )

                slicer = subsample_vector_field(self.__axis, self.__cut, subsample)
                vector_nda_slice = vector_nda.__getitem__(slicer)

                vector_ax = vector_nda_slice[:, :, 2].T
                vector_cor = vector_nda_slice[:, :, 1].T
                vector_sag = vector_nda_slice[:, :, 0].T

                (vector_plot_x, vector_plot_y, vector_plot_z,) = reorientate_vector_field(
                    self.__axis,
                    vector_ax,
                    vector_cor,
                    vector_sag,
                    invert_field=invert_field,
                )

                plot_x_loc, plot_y_loc = vector_image_grid(self.__axis, vector_nda, subsample)

                if color_function == "perpendicular":
                    vector_color = vector_plot_z
                elif color_function == "magnitude":
                    vector_color = np.sqrt(
                        vector_plot_x**2 + vector_plot_y**2 + vector_plot_z**2
                    )

                if max_value is False:
                    max_value = vector_color.max()
                if min_value is False:
                    min_value = vector_color.min()

                sp_vector = ax.quiver(
                    plot_x_loc,
                    plot_y_loc,
                    vector_plot_x,
                    vector_plot_y,
                    vector_color,
                    cmap=colormap,
                    units="xy",
                    scale=1 / arrow_scale,
                    width=arrow_width,
                    minlength=0,
                    linewidth=1,
                    alpha=alpha,
                    clim=[min_value, max_value],
                )

                # if vector.show_colorbar:
                #     divider = make_axes_locatable(ax)
                #     cax = divider.append_axes("right", size="5%", pad=0.05)
                #     cbar = self.__figure.colorbar(sp_vector, cax=cax, orientation="vertical")
                #     cbar.set_label(vector.name)
                #     cbar.solids.set_alpha(1)

                #     f_x, f_y = self.__figure.get_size_inches()
                #     self.__figure.set_size_inches(f_x * 1.15, f_y)
                #     self.__figure.subplots_adjust(left=0, right=0.88, bottom=0, top=1)

            elif len(axes) >= 4:
                ax_ax, _, ax_cor, ax_sag = axes[:4]

                # set this variable for colorbar reference
                ax = ax_ax

                for plot_axes, im_axis, im_cut in zip(
                    (ax_ax, ax_cor, ax_sag), ("z", "y", "x"), self.__cut
                ):

                    if not hasattr(subsample, "__iter__"):
                        subsample = (subsample,) * 3

                    subsample_img = [
                        int(np.ceil((i / j))) for i, j in zip(subsample, image.GetSpacing()[::-1])
                    ]

                    slicer = subsample_vector_field(im_axis, im_cut, subsample_img)
                    vector_nda_slice = vector_nda.__getitem__(slicer)

                    vector_ax = vector_nda_slice[:, :, 2].T
                    vector_cor = vector_nda_slice[:, :, 1].T
                    vector_sag = vector_nda_slice[:, :, 0].T

                    (
                        vector_plot_x,
                        vector_plot_y,
                        vector_plot_z,
                    ) = reorientate_vector_field(im_axis, vector_ax, vector_cor, vector_sag)

                    plot_x_loc, plot_y_loc = vector_image_grid(im_axis, vector_nda, subsample_img)

                    if color_function == "perpendicular":
                        vector_color = vector_plot_z
                    elif color_function == "magnitude":
                        vector_color = np.sqrt(
                            vector_plot_x**2 + vector_plot_y**2 + vector_plot_z**2
                        )

                    if max_value is False:
                        max_value = vector_color.max()
                    if min_value is False:
                        min_value = vector_color.min()

                    sp_vector = plot_axes.quiver(
                        plot_x_loc,
                        plot_y_loc,
                        vector_plot_x,
                        vector_plot_y,
                        vector_color,
                        cmap=colormap,
                        units="xy",
                        scale=1 / arrow_scale,
                        width=arrow_width,
                        minlength=0,
                        linewidth=1,
                        clim=[min_value, max_value],
                    )

            if vector.show_colorbar:

                # divider = make_axes_locatable(ax_view)
                # cax = divider.append_axes("right", size="5%", pad=0.05)

                ax_box = ax.get_position(original=False)
                cbar_width = ax_box.width * 0.05  # 5% of axis width
                cbar_color = "black"

                if self.__axis == "ortho":
                    cax = self.__figure.add_axes(
                        (
                            ax_box.x1 + 0.02 + (cbar_width + 0.1) * vector_index,
                            ax_box.y0 * 1.025,
                            cbar_width,
                            ax_box.height - ax_box.y0 * 0.05,
                        )
                    )

                else:
                    cax = self.__figure.add_axes(
                        (
                            ax_box.x1 - 0.02 - (cbar_width + 0.1) * (vector_index + 1),
                            0.025,
                            cbar_width,
                            ax_box.height - ax_box.y1 * 0.05,
                        )
                    )

                    # check background values
                    if np.linalg.norm(colormap(0)[:3]) < 0.1:
                        # background is dark
                        cbar_color = "white"

                cbar = self.__figure.colorbar(sp_vector, cax=cax, orientation="vertical")

                # set color
                cbar.outline.set_edgecolor(color=cbar_color)
                cbar.ax.tick_params(color=cbar_color)
                cax.tick_params(axis="x", colors=cbar_color)
                cax.tick_params(axis="y", colors=cbar_color)

                cbar.set_label(vector.name, color=cbar_color)
                cbar.solids.set_alpha(1)

                if vector.discrete_levels:

                    if vector.mid_ticks:

                        delta_tick = (max_value - min_value) / vector.discrete_levels
                        cbar.set_ticks(
                            np.linspace(
                                min_value + delta_tick / 2,
                                max_value - delta_tick / 2,
                                vector.discrete_levels,
                            )
                        )
                        cbar.set_ticklabels(
                            np.linspace(min_value, max_value, vector.discrete_levels)
                        )

                    else:
                        cbar.set_ticks(
                            np.linspace(
                                min_value,
                                max_value,
                                vector.discrete_levels + 1,
                            )
                        )

    def draw_bounding_box_on_axes(self, ax, view, box, add_label=True):

        sag_0, cor_0, ax_0, sag_d, cor_d, ax_d = box.bounding_box

        from_points = None
        to_points = None

        if view == "z" or view == "ax":
            from_points = [sag_0, sag_0, sag_0 + sag_d, sag_0 + sag_d, sag_0]
            to_points = [cor_0, cor_0 + cor_d, cor_0 + cor_d, cor_0, cor_0]

        if view == "y" or view == "cor":
            from_points = [sag_0, sag_0 + sag_d, sag_0 + sag_d, sag_0, sag_0]
            to_points = [ax_0, ax_0, ax_0 + ax_d, ax_0 + ax_d, ax_0]

        if view == "x" or view == "sag":
            from_points = [cor_0, cor_0 + cor_d, cor_0 + cor_d, cor_0, cor_0]
            to_points = [ax_0, ax_0, ax_0 + ax_d, ax_0 + ax_d, ax_0]

        if from_points is None:
            logger.error("Invalid View")
            return

        ax.plot(
            from_points,
            to_points,
            lw=box.linewidth,
            c=box.color,
            label=box.name if add_label else None,
        )

    def _overlay_bounding_boxes(self):
        """Overlay bounding boxes onto existing figure

        Args:
            color (str|list|tuple, optional): Color of bounding box. Defaults to "r".
        """

        for box in self.__bounding_boxes:

            # Test types of axes
            axes = self.__figure.axes[:4]
            if len(axes) < 4:
                self.draw_bounding_box_on_axes(axes[0], self.__axis, box)

            elif len(axes) == 4:
                ax_ax, _, ax_cor, ax_sag = axes

                self.draw_bounding_box_on_axes(ax_ax, "z", box)
                self.draw_bounding_box_on_axes(ax_cor, "y", box, add_label=False)
                self.draw_bounding_box_on_axes(ax_sag, "x", box, add_label=False)

    def _add_legend(self):
        """Add a legend to the visualisation"""

        if len(self.__figure.axes) >= 4:

            ax_ax = self.__figure.axes[0]
            ax_ax_position = ax_ax.get_position()
            y_pos_legend = (ax_ax_position.ymax + ax_ax_position.ymin) / 2

            if self.__show_legend:
                if len(self.__figure.axes) >= 5:
                    # There is a colorbar or more - right align instead

                    cbar_axes = self.__figure.axes[4:]

                    cbar_axes_label_pos_list = [
                        self.__figure.transFigure.inverted().transform(
                            i.yaxis.get_label().get_position()
                        )[0]
                        for i in cbar_axes
                    ]

                    max_xpos = max(cbar_axes_label_pos_list)

                    x_pos_legend = max_xpos + 0.025

                    approx_font_scaling = self.__figure_size / (
                        len(self.__contours) + len(self.__bounding_boxes)
                    )

                    plt.figlegend(
                        loc="center left",
                        bbox_to_anchor=(x_pos_legend, y_pos_legend),
                        fontsize=min([10, 16 * approx_font_scaling]),
                    )
                else:
                    # Left align to axial slice
                    ax_ax_position = ax_ax.get_position()
                    x_pos_legend = ax_ax_position.xmax + 0.05

                    approx_font_scaling = self.__figure_size / (
                        len(self.__contours) + len(self.__bounding_boxes)
                    )

                    plt.figlegend(
                        loc="center left",
                        bbox_to_anchor=(x_pos_legend, y_pos_legend),
                        fontsize=min([10, 16 * approx_font_scaling]),
                        ncol=1,  #!TODO modify this for large numbers of contours
                    )

        else:
            # these is probably only one axis
            if self.__show_legend:
                ax = self.__figure.axes[0]
                ax_position = ax.get_position()

                # top left corner is the default
                y_pos_legend = 0.975 * ax_position.ymax
                x_pos_legend = 0.025 * ax_position.xmax

                approx_font_scaling = self.__figure_size / (
                    len(self.__contours) + len(self.__bounding_boxes)
                )

                plt.figlegend(
                    loc="upper left",
                    bbox_to_anchor=(x_pos_legend, y_pos_legend),
                    fontsize=min([10, 16 * approx_font_scaling]),
                )



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from flask import Flask
import logging

logger = logging.getLogger(__name__)
import pydicom

from pymedphys._dicom.connect.listen import DicomListener


class Algorithm:
    def __init__(self, name, function, default_settings):
        self.name = name
        self.function = function
        self.default_settings = default_settings

    def settings_to_json(self):
        return json.dumps(self.default_settings, indent=4)


class FlaskApp(Flask):
    """
    Custom Flask App
    """

    algorithms = {}
    celery_started = False
    beat_started = False
    dicom_listener_port = 7777
    dicom_listener_aetitle = "PLATIPY_SERVICE"

    api = None  # Holds reference to api for extensibility

    def register(self, name, default_settings=None):
        def decorator(f):
            self.algorithms.update({name: Algorithm(name, f, default_settings)})
            return f

        return decorator

    def run(
        self,
        host=None,
        port=None,
        debug=None,
        dicom_listener_port=7777,
        dicom_listener_aetitle="PLATIPY_SERVICE",
        load_dotenv=True,
        **options
    ):

        logger.info("Starting APP!")

        self.dicom_listener_port = dicom_listener_port
        self.dicom_listener_aetitle = dicom_listener_aetitle

        self.run_dicom_listener(dicom_listener_port, dicom_listener_aetitle)

        super().run(
            host=host,
            port=port,
            debug=debug,
            load_dotenv=load_dotenv,
            use_reloader=False,
            **options
        )

    def run_dicom_listener(self, listen_port=None, listen_ae_title=None):
        """
        Background task that listens at a specific port for incoming dicom series
        """

        from .models import Dataset, DataObject
        from . import db

        if listen_port is None:
            listen_port = self.dicom_listener_port

        if listen_ae_title is None:
            listen_ae_title = self.dicom_listener_aetitle

        logger.info(
            "Starting Dicom Listener on port: %s with AE Title: %s",
            listen_port,
            listen_ae_title,
        )

        def series_recieved(dicom_path):
            logger.info("Series Recieved at path: %s", dicom_path)

            # Get the SeriesUID
            series_uid = None
            for f in os.listdir(dicom_path):
                f = os.path.join(dicom_path, f)

                try:
                    d = pydicom.read_file(f)
                    series_uid = d.SeriesInstanceUID
                except Exception as e:
                    logger.debug("No Series UID in: %s", f)
                    logger.debug(e)

            if series_uid:
                logger.info("Image Series UID: %s", series_uid)
            else:
                logger.error("Series UID could not be determined... Stopping")
                return

            # Find the data objects with the given series UID and update them
            dos = DataObject.query.filter_by(series_instance_uid=series_uid).all()

            if len(dos) == 0:
                logger.error("No Data Object found with Series UID: %s ... Stopping", series_uid)
                return

            for do in dos:

                do.is_fetched = True
                do.path = str(dicom_path)
                db.session.commit()

        try:
            dicom_listener = DicomListener(
                host="0.0.0.0",
                port=listen_port,
                ae_title=listen_ae_title,
                on_released_callback=series_recieved,
            )

            dicom_listener.start()

        except Exception as e:
            logger.error("Listener Error: %s", e)



# Copyright 2021 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pydicom
import SimpleITK as sitk


def convert_rtdose(dcm_dose, force=False, dose_output_path=None):
    """Convert DICOM RTDose to SimpleITK image, saving as NiFTI if needed.

    Args:
        dcm_dose (str|path): Path to DICOM dose file
        force (bool): force flag to pass to pydicom
        dose_output_path (str|path, optional): If set, NiFTI file will be written. Defaults to
            None.

    Returns:
        SimpleITK.Image: The dose grid as a SimpleITK image
    """

    ds = pydicom.read_file(dcm_dose, force=force)
    dose = sitk.ReadImage(str(dcm_dose))
    dose = sitk.Cast(dose, sitk.sitkFloat32)
    dose = dose * ds.DoseGridScaling

    if dose_output_path is not None:
        sitk.WriteImage(dose, str(dose_output_path))

    return dose



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=redefined-outer-name

import tempfile
from pathlib import Path
import logging

import pytest

import SimpleITK as sitk
import numpy as np

from platipy.imaging.generation.image import insert_sphere

from platipy.imaging.projects.cardiac.run import (
    run_cardiac_segmentation,
    CARDIAC_SETTINGS_DEFAULTS,
)
logger = logging.getLogger(__name__)


@pytest.fixture
def cardiac_data():
    """Generates the data needed for the cardiac tests

    Returns:
        dict -- Data for cardiac test
    """

    # Generate 5 pseudo CT images and the wholeheart masks
    data = {}
    for i in range(5):

        case_id = str(i + 1).zfill(3)

        ct_arr = np.ones((60, 128, 128)) * -1000
        mask_arr = np.zeros((60, 128, 128))
        submask_arr = np.zeros((60, 128, 128))

        ct_arr = insert_sphere(ct_arr, sp_radius=25, sp_centre=(30 + i, 64 + i, 64))
        mask_arr = insert_sphere(mask_arr, sp_radius=25, sp_centre=(30 + i, 64 + i, 64))
        submask_arr = insert_sphere(submask_arr, sp_radius=5, sp_centre=(30 + i, 60 + i, 60))

        ct = sitk.GetImageFromArray(ct_arr)
        ct.SetSpacing((0.9 + i * 0.01, 0.9 + i * 0.01, 2.5 + i * 0.01))
        ct.SetOrigin((320, -52, 60))

        mask = sitk.GetImageFromArray(mask_arr)
        mask.CopyInformation(ct)
        mask = sitk.Cast(mask, sitk.sitkUInt8)

        submask = sitk.GetImageFromArray(submask_arr)
        submask.CopyInformation(ct)
        submask = sitk.Cast(submask, sitk.sitkUInt8)

        data[case_id] = {"CT": ct, "WHOLEHEART": mask, "SUBSTRUCTURE": submask}

    return data


def test_cardiac_service(cardiac_data):
    """An end-to-end test to check that the cardiac service is running as expected"""

    with tempfile.TemporaryDirectory() as working_dir:

        working_path = Path(working_dir)

        # Save off data
        cases = list(cardiac_data.keys())
        for case in cardiac_data:

            ct_path = working_path.joinpath(f"Case_{case}", "Images", f"Case_{case}_CROP.nii.gz")
            ct_path.parent.mkdir(parents=True, exist_ok=True)
            mask_path = working_path.joinpath(
                f"Case_{case}", "Structures", f"Case_{case}_WHOLEHEART_CROP.nii.gz"
            )
            mask_path.parent.mkdir(parents=True, exist_ok=True)

            sitk.WriteImage(cardiac_data[case]["CT"], str(ct_path))
            sitk.WriteImage(cardiac_data[case]["WHOLEHEART"], str(mask_path))

        # Prepare algorithm settings
        test_settings = CARDIAC_SETTINGS_DEFAULTS
        test_settings["atlas_settings"]["atlas_id_list"] = cases[:-1]
        test_settings["atlas_settings"]["atlas_path"] = str(working_path)
        test_settings["atlas_settings"]["atlas_structure_list"] = ["WHOLEHEART"]
        test_settings["atlas_settings"]["auto_crop_atlas"] = False
        test_settings["deformable_registration_settings"]["resolution_staging"] = [8, 4, 2]
        test_settings["deformable_registration_settings"]["iteration_staging"] = [5, 5, 5]
        test_settings["deformable_registration_settings"]["smoothing_sigmas"] = [0, 0, 0]
        test_settings["deformable_registration_settings"]["default_value"] = -1000
        test_settings["iar_settings"]["reference_structure"] = None
        test_settings["label_fusion_settings"]["optimal_threshold"] = {"WHOLEHEART": 0.5}
        test_settings["vessel_spline_settings"]["vessel_name_list"] = []
        test_settings["vessel_spline_settings"]["vessel_radius_mm_dict"] = {}
        test_settings["vessel_spline_settings"]["scan_direction_dict"] = {}
        test_settings["vessel_spline_settings"]["stop_condition_type_dict"] = {}
        test_settings["vessel_spline_settings"]["stop_condition_value_dict"] = {}
        test_settings["postprocessing_settings"]["run_postprocessing"] = False
        test_settings["geometric_segmentation_settings"]["run_geometric_algorithms"] = False

        test_settings["rigid_settings"] = {
            "shrink_factors": [2, 1],
            "smooth_sigmas": [0, 0],
            "sampling_rate": 0.75,
            "default_value": -1024,
            "number_of_iterations": 5,
            "final_interp": sitk.sitkBSpline,
            "metric": "mean_squares",
            "optimiser": "gradient_descent_line_search",
        }

        # Run the service function.
        infer_case = cases[-1]

        output, _ = run_cardiac_segmentation(
            cardiac_data[infer_case]["CT"], settings=test_settings
        )

        # Check we have a WHOLEHEART structure
        assert "WHOLEHEART" in output

        # Check the result is similar to the GT

        label_overlap_filter = sitk.LabelOverlapMeasuresImageFilter()
        auto_mask = output["WHOLEHEART"]
        gt_mask = sitk.Cast(cardiac_data[infer_case]["WHOLEHEART"], auto_mask.GetPixelID())
        label_overlap_filter.Execute(auto_mask, gt_mask)
        assert label_overlap_filter.GetDiceCoefficient() > 0.99


def test_cardiac_structure_guided_service(cardiac_data):
    """
    An end-to-end test to check that the cardiac structure guided service is running as expected
    """

    with tempfile.TemporaryDirectory() as working_dir:

        working_path = Path(working_dir)

        # Save off data
        cases = list(cardiac_data.keys())
        for case in cardiac_data:

            ct_path = working_path.joinpath(f"Case_{case}", "Images", f"Case_{case}_CROP.nii.gz")
            ct_path.parent.mkdir(parents=True, exist_ok=True)
            mask_path = working_path.joinpath(
                f"Case_{case}", "Structures", f"Case_{case}_WHOLEHEART_CROP.nii.gz"
            )
            mask_path.parent.mkdir(parents=True, exist_ok=True)

            substructure_path = working_path.joinpath(
                f"Case_{case}", "Structures", f"Case_{case}_SUBSTRUCTURE_CROP.nii.gz"
            )

            sitk.WriteImage(cardiac_data[case]["CT"], str(ct_path))
            sitk.WriteImage(cardiac_data[case]["WHOLEHEART"], str(mask_path))
            sitk.WriteImage(cardiac_data[case]["SUBSTRUCTURE"], str(substructure_path))

        # Prepare algorithm settings
        test_settings = CARDIAC_SETTINGS_DEFAULTS
        test_settings["atlas_settings"]["atlas_id_list"] = cases[:-1]
        test_settings["atlas_settings"]["atlas_path"] = str(working_path)
        test_settings["atlas_settings"]["atlas_structure_list"] = ["WHOLEHEART", "SUBSTRUCTURE"]
        test_settings["atlas_settings"]["auto_crop_atlas"] = False
        test_settings["atlas_settings"]["guide_structure_name"] = "WHOLEHEART"
        test_settings["deformable_registration_settings"]["resolution_staging"] = [8, 4, 2]
        test_settings["deformable_registration_settings"]["iteration_staging"] = [5, 5, 5]
        test_settings["deformable_registration_settings"]["smoothing_sigmas"] = [0, 0, 0]
        test_settings["deformable_registration_settings"]["default_value"] = -1000
        test_settings["structure_guided_registration_settings"]["iteration_staging"] = [5, 5, 5]
        test_settings["structure_guided_registration_settings"]["resolution_staging"] = [6, 3, 1.5]
        test_settings["iar_settings"]["reference_structure"] = None
        test_settings["label_fusion_settings"]["optimal_threshold"] = {
            "WHOLEHEART": 0.5,
            "SUBSTRUCTURE": 0.5,
        }
        test_settings["vessel_spline_settings"]["vessel_name_list"] = []
        test_settings["vessel_spline_settings"]["vessel_radius_mm_dict"] = {}
        test_settings["vessel_spline_settings"]["scan_direction_dict"] = {}
        test_settings["vessel_spline_settings"]["stop_condition_type_dict"] = {}
        test_settings["vessel_spline_settings"]["stop_condition_value_dict"] = {}
        test_settings["postprocessing_settings"]["run_postprocessing"] = False
        test_settings["geometric_segmentation_settings"]["run_geometric_algorithms"] = False

        test_settings["rigid_settings"] = {
            "shrink_factors": [2, 1],
            "smooth_sigmas": [0, 0],
            "sampling_rate": 0.75,
            "default_value": -1024,
            "number_of_iterations": 5,
            "final_interp": sitk.sitkBSpline,
            "metric": "mean_squares",
            "optimiser": "gradient_descent_line_search",
        }

        # Run the service function.
        infer_case = cases[-1]

        output, _ = run_cardiac_segmentation(
            cardiac_data[infer_case]["CT"],
            cardiac_data[infer_case]["WHOLEHEART"],
            settings=test_settings,
        )
        # Check we have a WHOLEHEART structure
        assert "WHOLEHEART" in output

        # Check we have a SUBSTRUCTURE structure
        assert "SUBSTRUCTURE" in output

        # Check the result is similar to the GT

        label_overlap_filter = sitk.LabelOverlapMeasuresImageFilter()
        auto_mask = output["WHOLEHEART"]
        gt_mask = sitk.Cast(cardiac_data[infer_case]["WHOLEHEART"], auto_mask.GetPixelID())
        label_overlap_filter.Execute(auto_mask, gt_mask)
        logger.info("WH DSC %s", label_overlap_filter.GetDiceCoefficient())
        assert label_overlap_filter.GetDiceCoefficient() > 0.9

        auto_mask = output["SUBSTRUCTURE"]
        gt_mask = sitk.Cast(cardiac_data[infer_case]["SUBSTRUCTURE"], auto_mask.GetPixelID())
        label_overlap_filter.Execute(auto_mask, gt_mask)
        logger.info("SS DSC %s", label_overlap_filter.GetDiceCoefficient())
        assert label_overlap_filter.GetDiceCoefficient() > 0.9



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import SimpleITK as sitk

from platipy.imaging.label.utils import get_com

from platipy.imaging.generation.image import insert_sphere_image

from platipy.imaging.utils.crop import crop_to_roi, label_to_roi


def get_closest_point_2d(reference_label, measurement_label):
    """Finds the point on "measurement_label" that is closest to "reference_label"

    Args:
        reference_label (SimpleITK.Image): The reference label, from which distance is measured.
        measurement_label (SimpleITK.Image): The measurement label. We evaluate the distance at
            every location in this volume.

    Returns:
        tuple: The closest point in 2 dimensions.
    """

    # Compute the distance map
    distancemap_2d = sitk.SignedMaurerDistanceMap(
        reference_label, squaredDistance=False, useImageSpacing=True
    )

    # Get the distance from the reference to each point in the measurement label
    arr_distancemap = sitk.GetArrayFromImage(distancemap_2d)
    arr_measurement = sitk.GetArrayFromImage(measurement_label)

    yloc, xloc = np.where(arr_measurement)
    distances = arr_distancemap[yloc, xloc]

    # Find where the distance is a minimum
    location_of_min = distances.argmin()
    yloc, xloc = yloc[location_of_min], xloc[location_of_min]

    return yloc, xloc


def geometric_sinoatrialnode(label_svc, label_ra, label_wholeheart, radius_mm=10):
    """Geometric definition of the cardiac sinoatrial node (SAN).
    This is largely inspired by Loap et al 2021 [https://doi.org/10.1016/j.prro.2021.02.002]

    Args:
        label_svc (SimpleITK.Image): The binary mask defining the superior vena cava.
        label_ra (SimpleITK.Image): The binary mask defining the right atrium.
        label_wholeheart (SimpleITK.Image): The binary mask defining the whole heart.
        radius_mm (int, optional): The radius of the SAN, in millimetres. Defaults to 10.

    Returns:
        SimpleITK.Image: A binary mask defining the SAN
    """

    # To speed up binary morphology operations we first crop all images
    template_img = 0 * label_wholeheart
    cb_size, cb_index = label_to_roi(
        (label_svc + label_ra + label_wholeheart) > 0, expansion_mm=(20, 20, 20)
    )

    label_svc = crop_to_roi(label_svc, cb_size, cb_index)
    label_ra = crop_to_roi(label_ra, cb_size, cb_index)
    label_wholeheart = crop_to_roi(label_wholeheart, cb_size, cb_index)

    arr_svc = sitk.GetArrayFromImage(label_svc)
    arr_ra = sitk.GetArrayFromImage(label_ra)

    # First, find the most inferior slice of the SVC
    # This defines the z location of the SAN
    inf_limit_svc = np.min(np.where(arr_svc)[0])

    # Now expand the SVC until it touches the RA on the inf slice
    overlap = 0
    dilate = 1
    dilate_ax = 0
    while overlap == 0:
        label_svc_dilate = sitk.BinaryDilate(label_svc, (dilate, dilate, dilate_ax))
        label_overlap = label_svc_dilate & label_ra
        overlap = sitk.GetArrayFromImage(label_overlap)[inf_limit_svc, :, :].sum()
        dilate += 1

        if dilate >= 3:
            arr_svc = sitk.GetArrayFromImage(label_svc_dilate)
            inf_limit_svc = np.min(np.where(arr_svc)[0])
            dilate_ax += 1

    # Locate the point on intersection
    intersect_loc = get_com(label_overlap)

    # Create an image with a single voxel of value 1 located at the point of intersection
    arr_intersect = arr_ra * 0
    arr_intersect[inf_limit_svc, intersect_loc[1], intersect_loc[2]] = 1
    label_intersect = sitk.GetImageFromArray(arr_intersect)
    label_intersect.CopyInformation(label_ra)

    # Define the locations greater than 10mm from the WH
    # Ensures the SAN doesn't extend outside the heart volume
    potential_san_region = sitk.BinaryErode(label_wholeheart, (10, 10, 0))

    # Find the point in this region closest to the intersection
    # First generate a distance map
    distancemap_san = sitk.SignedMaurerDistanceMap(
        label_intersect, squaredDistance=False, useImageSpacing=True
    )

    # Then get the distance from the intersection at all possible points
    arr_distancemap_san = sitk.GetArrayFromImage(distancemap_san)
    arr_potential_san_region = sitk.GetArrayFromImage(potential_san_region)

    yloc, xloc = np.where(arr_potential_san_region[inf_limit_svc, :, :])

    distances = arr_distancemap_san[inf_limit_svc, yloc, xloc]

    # Find where the distance is a minimum
    location_of_min = distances.argmin()

    # Now define the SAN location
    sphere_centre = (inf_limit_svc, yloc[location_of_min], xloc[location_of_min])

    # Generate an image
    label_san = insert_sphere_image(label_ra * 0, sp_radius=radius_mm, sp_centre=sphere_centre)

    # Finally, paste the label into the original image space
    label_san = sitk.Paste(
        template_img,
        label_san,
        label_san.GetSize(),
        (0, 0, 0),
        cb_index,
    )

    return label_san


def geometric_atrioventricularnode(label_la, label_lv, label_ra, label_rv, radius_mm=10):
    """Geometric definition of the cardiac atrioventricular node (AVN).
    This is largely inspired by Loap et al 2021 [https://doi.org/10.1016/j.prro.2021.02.002]

    Args:
        label_la (SimpleITK.Image): The binary mask defining the left atrium.
        label_lv (SimpleITK.Image): The binary mask defining the left ventricle.
        label_ra (SimpleITK.Image): The binary mask defining the right atrium.
        label_rv (SimpleITK.Image): The binary mask defining the right ventricle.
        radius_mm (float, optional): The radius of the AVN, in millimetres. Defaults to 10.

    Returns:
        SimpleITK.Image: A binary mask defining the AVN
    """

    # To speed up binary morphology operations we first crop all images
    template_img = 0 * label_ra
    cb_size, cb_index = label_to_roi(
        (label_la + label_lv + label_ra + label_rv) > 0, expansion_mm=(20, 20, 20)
    )

    label_la = crop_to_roi(label_la, cb_size, cb_index)
    label_lv = crop_to_roi(label_lv, cb_size, cb_index)
    label_ra = crop_to_roi(label_ra, cb_size, cb_index)
    label_rv = crop_to_roi(label_rv, cb_size, cb_index)

    # First, find the most inferior slice of the left atrium
    arr_la = sitk.GetArrayFromImage(label_la)
    inf_limit_la = np.min(np.where(arr_la)[0])

    # Now progress 1cm in the superior direction
    # This defines the slice of the AVN centre
    slice_loc = int(inf_limit_la + 10 / label_la.GetSpacing()[2])

    # Create 2D images at this slice location
    label_la_2d = label_la[:, :, slice_loc]
    label_lv_2d = label_lv[:, :, slice_loc]
    label_ra_2d = label_ra[:, :, slice_loc]
    label_rv_2d = label_rv[:, :, slice_loc]

    # We now iteratively erode the structures to ensure they do not overlap
    # This ensures we can measure the closest point without any errors
    # LEFT ATRIUM
    overlap = 1
    erode = 1
    while overlap > 0:
        label_lv_2d = sitk.BinaryErode(label_lv_2d, (erode, erode))
        label_overlap = label_lv_2d & label_la_2d
        overlap = sitk.GetArrayFromImage(label_overlap).sum()
        erode += 1

    # LEFT ATRIUM
    overlap = 0
    erode = 1
    while overlap > 0:
        label_la_2d = sitk.BinaryErode(label_la_2d, (erode, erode))
        label_overlap = label_la_2d & label_ra_2d
        overlap = sitk.GetArrayFromImage(label_overlap).sum()
        erode += 1

    # RIGHT ATRIUM
    overlap = 0
    erode = 1
    while overlap > 0:
        label_ra_2d = sitk.BinaryErode(label_ra_2d, (erode, erode))
        label_overlap = label_ra_2d & label_rv_2d
        overlap = sitk.GetArrayFromImage(label_overlap).sum()
        erode += 1

    # RIGHT VENTRICLE
    overlap = 0
    erode = 1
    while overlap > 0:
        label_rv_2d = sitk.BinaryErode(label_rv_2d, (erode, erode))
        label_overlap = label_rv_2d & label_lv_2d
        overlap = sitk.GetArrayFromImage(label_overlap).sum()
        erode += 1

    # Measure closest points
    y_la, x_la = get_closest_point_2d(label_rv_2d, label_la_2d)
    y_lv, x_lv = get_closest_point_2d(label_ra_2d, label_lv_2d)
    y_ra, x_ra = get_closest_point_2d(label_lv_2d, label_ra_2d)
    y_rv, x_rv = get_closest_point_2d(label_la_2d, label_rv_2d)

    # Take the arithmetic mean
    x_location = np.mean((x_la, x_lv, x_ra, x_rv), dtype=int)
    y_location = np.mean((y_la, y_lv, y_ra, y_rv), dtype=int)

    # Now define the AVN location
    sphere_centre = (slice_loc, y_location, x_location)

    # Generate an image
    label_avn = insert_sphere_image(label_ra * 0, sp_radius=radius_mm, sp_centre=sphere_centre)

    # Finally, paste the label into the original image space
    label_avn = sitk.Paste(
        template_img,
        label_avn,
        label_avn.GetSize(),
        (0, 0, 0),
        cb_index,
    )

    return label_avn



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import warnings

import logging

import numpy as np
import SimpleITK as sitk

logger = logging.getLogger(__name__)

try:
    import vtk
except ImportError as e:
    raise ImportError("Unable to import VTK library. To use cardiac segmentation functionality be "
                      "sure to install appropriate extras: pip install platipy[cardiac]") from e

from vtk.util.numpy_support import vtk_to_numpy

def com_from_image_list(
    sitk_image_list,
    condition_type="count",
    condition_value=0,
    scan_direction="z",
):
    """
    Input: list of SimpleITK images
           minimum total slice area required for the tube to be inserted at that slice
           scan direction: x = sagittal, y=coronal, z=axial
    Output: mean centre of mass positions, with shape (NumSlices, 2)
    Note: positions are converted into image space by default
    """
    if scan_direction.lower() == "x":
        logger.debug("Scanning in sagittal direction")
        com_z_list = []
        com_y_list = []
        weight_list = []
        count_list = []

        reference_image = sitk_image_list[0]
        reference_array = sitk.GetArrayFromImage(reference_image)
        z, y = np.mgrid[0 : reference_array.shape[0] : 1, 0 : reference_array.shape[1] : 1]

        with np.errstate(divide="ignore", invalid="ignore"):
            for sitk_image in sitk_image_list:
                volume_array = sitk.GetArrayFromImage(sitk_image)
                com_z = 1.0 * (z[:, :, np.newaxis] * volume_array).sum(axis=(1, 0))
                com_y = 1.0 * (y[:, :, np.newaxis] * volume_array).sum(axis=(1, 0))

                weights = np.sum(volume_array, axis=(1, 0))
                weight_list.append(weights)

                count_list.append(np.any(volume_array, axis=(1, 0)))

                com_z /= 1.0 * weights
                com_y /= 1.0 * weights

                com_z_list.append(com_z)
                com_y_list.append(com_y)

        with warnings.catch_warnings():
            """
            It's fairly likely some slices have just np.NaN values
            It raises a warning but we can suppress it here
            """
            warnings.simplefilter("ignore", category=RuntimeWarning)
            mean_com_z_list = np.nanmean(com_z_list, axis=0)
            mean_com_y_list = np.nanmean(com_y_list, axis=0)
            if condition_type.lower() == "area":
                mean_com = (
                    np.dstack((mean_com_z_list, mean_com_y_list))[0]
                    * np.array((np.sum(weight_list, axis=0) > (condition_value),) * 2).T
                )
            elif condition_type.lower() == "count":
                mean_com = (
                    np.dstack((mean_com_z_list, mean_com_y_list))[0]
                    * np.array((np.sum(count_list, axis=0) > (condition_value),) * 2).T
                )
            else:
                raise ValueError("Invalid condition type, please select from 'area' or 'count'.")

        point_array = []
        for index, com in enumerate(mean_com):
            if np.all(np.isfinite(com)):
                if np.all(com > 0):
                    point_array.append(
                        reference_image.TransformIndexToPhysicalPoint(
                            (index, int(com[1]), int(com[0]))
                        )
                    )

        return point_array

    elif scan_direction.lower() == "z":
        logger.debug("Scanning in axial direction")
        com_x_list = []
        com_y_list = []
        weight_list = []
        count_list = []

        reference_image = sitk_image_list[0]
        reference_array = sitk.GetArrayFromImage(reference_image)
        x, y = np.mgrid[0 : reference_array.shape[1] : 1, 0 : reference_array.shape[2] : 1]

        with np.errstate(divide="ignore", invalid="ignore"):
            for sitk_image in sitk_image_list:
                volume_array = sitk.GetArrayFromImage(sitk_image)
                com_x = 1.0 * (x * volume_array).sum(axis=(1, 2))
                com_y = 1.0 * (y * volume_array).sum(axis=(1, 2))

                weights = np.sum(volume_array, axis=(1, 2))
                weight_list.append(weights)

                count_list.append(np.any(volume_array, axis=(1, 2)))

                com_x /= 1.0 * weights
                com_y /= 1.0 * weights

                com_x_list.append(com_x)
                com_y_list.append(com_y)

        with warnings.catch_warnings():
            """
            It's fairly likely some slices have just np.NaN values - it raises a warning but we
            can suppress it here
            """
            warnings.simplefilter("ignore", category=RuntimeWarning)
            mean_com_x_list = np.nanmean(com_x_list, axis=0)
            mean_com_y_list = np.nanmean(com_y_list, axis=0)
            if condition_type.lower() == "area":
                mean_com = (
                    np.dstack((mean_com_x_list, mean_com_y_list))[0]
                    * np.array((np.sum(weight_list, axis=0) > (condition_value),) * 2).T
                )
            elif condition_type.lower() == "count":
                mean_com = (
                    np.dstack((mean_com_x_list, mean_com_y_list))[0]
                    * np.array((np.sum(count_list, axis=0) > (condition_value),) * 2).T
                )
            else:
                print("Invalid condition type, please select from 'area' or 'count'.")
                quit()

        point_array = []
        for index, com in enumerate(mean_com):
            if np.all(np.isfinite(com)):
                if np.all(com > 0):
                    point_array.append(
                        reference_image.TransformIndexToPhysicalPoint(
                            (int(com[1]), int(com[0]), index)
                        )
                    )

        return point_array


def tube_from_com_list(com_list, radius):
    """
    Input: image-space positions along the tube centreline.
    Output: VTK tube
    Note: positions do not have to be continuous - the tube is interpolated in real space
    """
    points = vtk.vtkPoints()
    for i, pt in enumerate(com_list):
        points.InsertPoint(i, pt[0], pt[1], pt[2])

    # Fit a spline to the points
    logger.debug("Fitting spline")
    spline = vtk.vtkParametricSpline()
    spline.SetPoints(points)

    function_source = vtk.vtkParametricFunctionSource()
    function_source.SetParametricFunction(spline)
    function_source.SetUResolution(10 * points.GetNumberOfPoints())
    function_source.Update()

    # Generate the radius scalars
    tube_radius = vtk.vtkDoubleArray()
    n = function_source.GetOutput().GetNumberOfPoints()
    tube_radius.SetNumberOfTuples(n)
    tube_radius.SetName("TubeRadius")
    for i in range(n):
        # We can set the radius based on the given propagated segmentations in that slice?
        # Typically segmentations are elliptical, this could be an issue so for now a constant
        # radius is used
        tube_radius.SetTuple1(i, radius)

    # Add the scalars to the polydata
    tube_poly_data = vtk.vtkPolyData()
    tube_poly_data = function_source.GetOutput()
    tube_poly_data.GetPointData().AddArray(tube_radius)
    tube_poly_data.GetPointData().SetActiveScalars("TubeRadius")

    # Create the tubes
    tuber = vtk.vtkTubeFilter()
    tuber.SetInputData(tube_poly_data)
    tuber.SetNumberOfSides(50)
    tuber.SetVaryRadiusToVaryRadiusByAbsoluteScalar()
    tuber.Update()

    return tuber


def write_vtk_tube_to_file(tube, filename):
    """
    Input: VTK tube
    Output: exit success
    Note: format is XML VTP
    """
    print("Writing tube to polydata file (VTP)")
    poly_data_writer = vtk.vtkXMLPolyDataWriter()
    poly_data_writer.SetInputData(tube.GetOutput())

    poly_data_writer.SetFileName(filename)
    poly_data_writer.SetCompressorTypeToNone()
    poly_data_writer.SetDataModeToAscii()
    s = poly_data_writer.Write()

    return s


def simpleitk_image_from_vtk_tube(tube, sitk_reference_image):
    """
    Input: VTK tube, referenceImage (used for spacing, etc.)
    Output: SimpleITK image
    Note: Uses binary output (background 0, foreground 1)
    """
    size = list(sitk_reference_image.GetSize())
    origin = list(sitk_reference_image.GetOrigin())
    spacing = list(sitk_reference_image.GetSpacing())
    ncomp = sitk_reference_image.GetNumberOfComponentsPerPixel()

    # convert the SimpleITK image to a numpy array
    arr = sitk.GetArrayFromImage(sitk_reference_image).transpose(2, 1, 0).flatten()

    # send the numpy array to VTK with a vtkImageImport object
    data_importer = vtk.vtkImageImport()

    data_importer.CopyImportVoidPointer(arr, len(arr))
    data_importer.SetDataScalarTypeToUnsignedChar()
    data_importer.SetNumberOfScalarComponents(ncomp)

    # Set the new VTK image's parameters
    data_importer.SetDataExtent(0, size[0] - 1, 0, size[1] - 1, 0, size[2] - 1)
    data_importer.SetWholeExtent(0, size[0] - 1, 0, size[1] - 1, 0, size[2] - 1)
    data_importer.SetDataOrigin(origin)
    data_importer.SetDataSpacing(spacing)

    data_importer.Update()

    vtk_reference_image = data_importer.GetOutput()

    # fill the image with foreground voxels:
    inval = 1
    outval = 0
    vtk_reference_image.GetPointData().GetScalars().Fill(inval)

    logger.debug("Using polydaya to generate stencil.")
    pol2stenc = vtk.vtkPolyDataToImageStencil()
    pol2stenc.SetTolerance(0.5)  # points within 0.5 voxels are included
    pol2stenc.SetInputConnection(tube.GetOutputPort())
    pol2stenc.SetOutputOrigin(vtk_reference_image.GetOrigin())
    pol2stenc.SetOutputSpacing(vtk_reference_image.GetSpacing())
    pol2stenc.SetOutputWholeExtent(vtk_reference_image.GetExtent())
    pol2stenc.Update()

    logger.debug("using stencil to generate image.")
    imgstenc = vtk.vtkImageStencil()
    imgstenc.SetInputData(vtk_reference_image)
    imgstenc.SetStencilConnection(pol2stenc.GetOutputPort())
    imgstenc.ReverseStencilOff()
    imgstenc.SetBackgroundValue(outval)
    imgstenc.Update()

    logger.debug("Generating SimpleITK image.")
    final_image = imgstenc.GetOutput()
    final_array = final_image.GetPointData().GetScalars()
    final_array = vtk_to_numpy(final_array).reshape(sitk_reference_image.GetSize()[::-1])
    logger.debug("Volume = %.3f mm^3", final_array.sum()*sum(spacing))
    final_image_sitk = sitk.GetImageFromArray(final_array)
    final_image_sitk.CopyInformation(sitk_reference_image)

    return final_image_sitk


def convert_simpleitk_to_vtk(img):
    """Converts from SimpleITK to VTK representation

    Args:
        img (SimpleITK.Image): The input image.

    Returns:
        The VTK image.
    """
    size = list(img.GetSize())
    origin = list(img.GetOrigin())
    spacing = list(img.GetSpacing())
    ncomp = img.GetNumberOfComponentsPerPixel()

    # convert the SimpleITK image to a numpy array
    arr = sitk.GetArrayFromImage(img).transpose(2, 1, 0).flatten()
    arr_string = arr.tostring()

    # send the numpy array to VTK with a vtkImageImport object
    data_importer = vtk.vtkImageImport()

    data_importer.CopyImportVoidPointer(arr_string, len(arr_string))
    data_importer.SetDataScalarTypeToUnsignedChar()
    data_importer.SetNumberOfScalarComponents(ncomp)

    # Set the new VTK image's parameters
    data_importer.SetDataExtent(0, size[0] - 1, 0, size[1] - 1, 0, size[2] - 1)
    data_importer.SetWholeExtent(0, size[0] - 1, 0, size[1] - 1, 0, size[2] - 1)
    data_importer.SetDataOrigin(origin)
    data_importer.SetDataSpacing(spacing)

    data_importer.Update()

    vtk_image = data_importer.GetOutput()
    return vtk_image


def vessel_spline_generation(
    reference_image,
    atlas_set,
    vessel_name_list,
    vessel_radius_mm_dict,
    stop_condition_type_dict,
    stop_condition_value_dict,
    scan_direction_dict,
    atlas_label="DIR",
):
    """Generates a splined vessel from a list of binary masks.


    Args:
        reference_image (SimpleITK.Image): The reference image to copy information from.
        atlas_set (dict): A dictionary conforming to the following format:
            { atlas_id:
                {
                    LABEL:
                        {
                            structure_a: SimpleITK.Image,
                            structure_b: SimpleITK.Image,
                            structure_c: SimpleITK.Image,
                            ...
                        }
                }
                ...
            }
            where LABEL should be passed as the atlas_label argument, and atlas_id is some label
            for each atlas.
        vessel_name_list (list): The list of vessels to generate splines for.
        vessel_radius_mm_dict (list): A dictionary specifying the radius for each vessel.
        stop_condition_type_dict (dict): A dictionary specifying the stopping condition for each
            vessel. Available options are "count" - stopping at a particular number of atlases, or
            "area" - stopping at a particular total area (in square pixels). Recommendation is to
            use "count" with a stop_condition_value (see below) of ~1/3 the number of atlases.
        stop_condition_value_dict (dict):  A dictionary specifying the stopping value for each
            vessel.
        scan_direction_dict (dict): A dictionary specifying the direction to spline each vessel.
            "x": sagittal, "y": coronal, "z": axial.
        atlas_label (str, optional): The atlas label. Defaults to "DIR".

    Returns:
        dict: The output dictionary, with keys as the vessel names and values as the binary labels
            defining the splined vessels
    """
    splined_vessels = {}

    if isinstance(vessel_name_list, str):
        vessel_name_list = [vessel_name_list]

    for vessel_name in vessel_name_list:

        # We must set the image direction to identity
        # This is because it is not possible to modify VTK Image directions
        # This may get fixed in a future VTK version

        initial_image_direction = reference_image.GetDirection()

        image_list = []
        for i in atlas_set.keys():
            try:
                image_list.append(atlas_set[i][atlas_label][vessel_name])
            except:
                logger.warning("No match for ID=%s, label=%s, vessel=%s", i, atlas_label, vessel_name)

        if len(image_list) == 0:
            logger.warning("No structures found for vessel with name %s!", vessel_name)
            continue

        for im in image_list:
            im.SetDirection((1, 0, 0, 0, 1, 0, 0, 0, 1))

        vessel_radius = vessel_radius_mm_dict[vessel_name]
        stop_condition_type = stop_condition_type_dict[vessel_name]
        stop_condition_value = stop_condition_value_dict[vessel_name]
        scan_direction = scan_direction_dict[vessel_name]

        point_array = com_from_image_list(
            image_list,
            condition_type=stop_condition_type,
            condition_value=stop_condition_value,
            scan_direction=scan_direction,
        )
        tube = tube_from_com_list(point_array, radius=vessel_radius)

        reference_image = image_list[0]

        vessel_delineation = simpleitk_image_from_vtk_tube(tube, reference_image)

        vessel_delineation.SetDirection(initial_image_direction)

        splined_vessels[vessel_name] = vessel_delineation

        # We also have to reset the direction to whatever it was
        # This is because SimpleITK doesn't use deep copying
        # And it isn't necessary here as we can save some sweet, sweet memory
        for im in image_list:
            im.SetDirection(initial_image_direction)

    return splined_vessels


# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import SimpleITK as sitk
import pandas as pd


def calculate_d_mean(dose_grid, label):
    """Calculate the mean dose of a structure

    Args:
        dose_grid (SimpleITK.Image): The dose grid.
        label (SimpleITK.Image): The (binary) label defining a structure.

    Returns:
        float: The mean dose in Gy.
    """

    dose_grid = sitk.Resample(dose_grid, label, sitk.Transform(), sitk.sitkLinear)
    dose_array = sitk.GetArrayFromImage(dose_grid)
    mask_array = sitk.GetArrayFromImage(label)

    return dose_array[mask_array > 0].mean()


def calculate_d_max(dose_grid, label):
    """Calculate the maximum dose of a structure

    Args:
        dose_grid (SimpleITK.Image): The dose grid.
        label (SimpleITK.Image): The (binary) label defining a structure.

    Returns:
        float: The maximum dose in Gy.
    """

    dose_grid = sitk.Resample(dose_grid, label, sitk.Transform(), sitk.sitkLinear)
    dose_array = sitk.GetArrayFromImage(dose_grid)
    mask_array = sitk.GetArrayFromImage(label)

    return dose_array[mask_array > 0].max()


def calculate_d_to_volume(dose_grid, label, volume, volume_in_cc=False):
    """Calculate the dose to a (relative) volume of the label

    Args:
        dose_grid (SimpleITK.Image): The dose grid.
        label (SimpleITK.Image): The (binary) label defining a structure.
        volume (float): The relative volume in %.
        volume_in_cc (bool, optional): Whether the volume is in cc (versus percent).
            Defaults to False.

    Returns:
        float: The dose to volume ratio.
    """

    dose_grid = sitk.Resample(dose_grid, label, sitk.Transform(), sitk.sitkLinear)
    dose_array = sitk.GetArrayFromImage(dose_grid)
    mask_array = sitk.GetArrayFromImage(label)

    if volume_in_cc:
        volume = (volume * 1000 / ((mask_array > 0).sum() * np.product(label.GetSpacing()))) * 100

    if volume > 100:
        volume = 100

    return np.percentile(dose_array[mask_array > 0], 100 - volume)


def calculate_v_receiving_dose(dose_grid, label, dose_threshold, relative=True):
    """Calculate the (relative) volume receiving a dose above a threshold

    Args:
        dose_grid (SimpleITK.Image): The dose grid.
        label (SimpleITK.Image): The (binary) label defining a structure.
        dose_threshold (float): The dose threshold in Gy.
        relative (bool, optional): If true results will be returned as relative volume, otherwise
            as volume in cc. Defaults to True.

    Returns:
        float: The (relative) volume receiving a dose above the threshold, as a percent.
    """

    dose_grid = sitk.Resample(dose_grid, label, sitk.Transform(), sitk.sitkLinear)
    dose_array = sitk.GetArrayFromImage(dose_grid)
    mask_array = sitk.GetArrayFromImage(label)

    dose_array_masked = dose_array[mask_array > 0]

    num_voxels = (mask_array > 0).sum()

    relative_volume = (dose_array_masked >= dose_threshold).sum() / num_voxels * 100
    if relative:
        return relative_volume

    total_volume = (mask_array > 0).sum() * np.product(label.GetSpacing()) / 1000

    return relative_volume * total_volume


def calculate_d_to_volume_for_labels(dose_grid, labels, volume, volume_in_cc=False):
    """Calculate the dose which x percent of the volume receives for a set of labels

    Args:
        dose_grid (SimpleITK.Image): The dose grid.
        labels (dict): A Python dictionary containing the label name as key and the SimpleITK.Image
          binary mask as value.
        volume (float|list): The relative volume (or list of volumes) in %.
        volume_in_cc (bool, optional): Whether the volume is in cc (versus percent).
            Defaults to False.

    Returns:
        pandas.DataFrame: Data frame with a row for each label containing the metric and value.
    """

    if not isinstance(volume, list):
        volume = [volume]

    metrics = []
    for label in labels:

        m = {"label": label}

        for v in volume:
            col_name = f"D{v}"
            if volume_in_cc:
                col_name = f"D{v}cc"

            m[col_name] = calculate_d_to_volume(
                dose_grid, labels[label], v, volume_in_cc=volume_in_cc
            )

        metrics.append(m)

    return pd.DataFrame(metrics)


def calculate_v_receiving_dose_for_labels(dose_grid, labels, dose_threshold, relative=True):
    """Get the volume (in cc) which receives x dose for a set of labels

    Args:
        dose_grid (SimpleITK.Image): The dose grid.
        labels (SimpleITK.Image): The (binary) label defining a structure.
        dose_threshold (float|list): The dose threshold (or list of thresholds) in Gy.
        relative (bool, optional): If true results will be returned as relative volume, otherwise
            as volume in cc. Defaults to True.

    Returns:
        pandas.DataFrame: Data frame with a row for each label containing the metric and value.
    """

    if not isinstance(dose_threshold, list):
        dose_threshold = [dose_threshold]

    metrics = []
    for label in labels:

        m = {"label": label}

        for dt in dose_threshold:

            metric_name = f"V{dt}"
            if dt - int(dt) == 0:
                metric_name = f"V{int(dt)}"

            m[metric_name] = calculate_v_receiving_dose(dose_grid, labels[label], dt, relative)

        metrics.append(m)

    return pd.DataFrame(metrics)



#!/usr/bin/env python

# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import click

import logging

logger = logging.getLogger(__name__)

from platipy.dicom.io.crawl import (
    process_dicom_directory,
)


@click.command()
@click.option(
    "--input_dir",
    "-i",
    required=True,
    type=click.Path(),
    help="Input DICOM directory. This should be at the same level as the parent field "
    "(default=PatientName).",
)
@click.option(
    "--output_dir",
    "-o",
    default="./",
    show_default=True,
    required=False,
    type=click.Path(),
    help="Output directory. A folder structure will be created at this location.",
)
@click.option(
    "--sort_by",
    "-b",
    default="PatientName",
    help="DICOM tag to sort at the highest level.",
    show_default=True,
)
@click.option(
    "--image_format",
    default="{parent_sorting_data}_{study_uid_index}_{Modality}_{image_desc}_{SeriesNumber}",
    help="Format for output images. There are three special options that can be used: "
    "parent_sorting_data (same as sort_by option), study_uid_index (a counter for distinct DICOM "
    "studies), image_desc (info from DICOM header, more nicely formatted). Additionally, any "
    "DICOM header tag can be used (e.g. Modality, SeriesNumber, AcquisitionData). Any DICOM "
    "header tag that doesn't exist will return a 0.",
    show_default=True,
)
@click.option(
    "--structure_format",
    default="{parent_sorting_data}_{study_uid_index}_{Modality}_{structure_name}",
    help="Format for output structures. Any of the options for images can be used, as "
    "well as: structure_name",
    show_default=True,
)
@click.option(
    "--dose_format",
    default="{parent_sorting_data}_{study_uid_index}_{DoseSummationType}",
    show_default=True,
    help="Format for output radiotherapy dose distributions.",
)
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    help="Overwrite files if they exist.",
    show_default=True,
)
@click.option(
    "--file_suffix",
    default=".nii.gz",
    help="Output file suffix. Defines the file type.",
    show_default=True,
)
@click.option(
    "--short_description",
    "-s",
    is_flag=True,
    default=False,
    show_default=True,
    help="Use less verbose descriptions for DICOM images.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    show_default=True,
    help="Print more information while running.",
)
def click_command(
    input_dir,
    output_dir,
    sort_by,
    image_format,
    structure_format,
    dose_format,
    overwrite,
    file_suffix,
    short_description,
    verbose,
):
    """
    DICOM DIRECTORY CRAWLER

    This tool makes it easier to bulk-convert DICOM files into other formats (default NifTI).
    There are quite a lot of options, but most do not need to be changed.

    You need to provide the input directory (-i), from which the crawler will recursively search
    through.

    You might also like to change the naming format (using --image_format and --structure_format).
    The default is quite long, better suited for datasets with lots of imaging for a single
    patient.

    Some examples:

      [simple] --image_format {parent_sorting_data}

      [compact]  --image_format {parent_sorting_data}_{study_uid_index}

    You can separate series using different values (--sort_by ).
    This would typically be PatientName, or PatientID, although any DICOM header tag is allowed.

    I hope you find this tool useful!

    If you have any feedback, let us know on github.com/pyplati/platipy

    """
    logger.info("########################")
    logger.info(" Running DICOM crawler ")
    logger.info("########################")

    process_dicom_directory(
        input_dir,
        parent_sorting_field=sort_by,
        output_image_name_format=image_format,
        output_structure_name_format=structure_format,
        output_dose_name_format=dose_format,
        return_extra=(not short_description),
        output_directory=output_dir,
        output_file_suffix=file_suffix,
        overwrite_existing_files=overwrite,
        write_to_disk=True,
        verbose=verbose,
    )

    logger.info("########################")
    logger.info(" DICOM crawler complete")
    logger.info("########################")


if __name__ == "__main__":
    click_command()  # pylint: disable=no-value-for-parameter
